from .analyzers.aa import OneAAStatAnalyzer
from .comparators import GroupDifference, GroupSizes
from .comparators import TTest, KSTest, Chi2Test
from .dataset import TargetRole, TreatmentRole
from .experiments.base import Experiment, OnRoleExperiment
from .ui.base import ExperimentShell
from .ui.homo import HomoOutput

HOMOGENEITY_TEST = Experiment(
    executors=[
        OnRoleExperiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole(), compare_by="groups"),
                GroupDifference(grouping_role=TreatmentRole(), compare_by="groups"),
                TTest(grouping_role=TreatmentRole(), compare_by="groups"),
                KSTest(grouping_role=TreatmentRole(), compare_by="groups"),
                Chi2Test(grouping_role=TreatmentRole(), compare_by="groups"),
            ],
            role=TargetRole(),
        ),
        OneAAStatAnalyzer(),
    ]
)


class HomogeneityTest(ExperimentShell):
    def __init__(self):
        super().__init__(
            experiment=HOMOGENEITY_TEST,
            output=HomoOutput(),
        )



from typing import Dict, Union, Any, Optional

from ..dataset import ExperimentData, Dataset
from ..experiments.base import Experiment
from ..reporters import Reporter
from ..utils import ID_SPLIT_SYMBOL
from ..utils.enums import RenameEnum


class Output:
    resume: Dataset
    _experiment_data: ExperimentData

    def __init__(
        self,
        resume_reporter: Reporter,
        additional_reporters: Optional[Dict[str, Reporter]] = None,
    ):
        self.resume_reporter = resume_reporter
        self.additional_reporters = additional_reporters or {}

    def _extract_by_reporters(self, experiment_data: ExperimentData):
        self.resume = self.resume_reporter.report(experiment_data)
        for attribute, reporter in self.additional_reporters.items():
            setattr(self, attribute, reporter.report(experiment_data))
        self._experiment_data = experiment_data

    @staticmethod
    def replace_splitters(
        data: Dataset, mode: RenameEnum = RenameEnum.columns
    ) -> Dataset:
        result = data
        if mode in (RenameEnum.all, RenameEnum.columns):
            result = result.rename(
                {c: c.replace(ID_SPLIT_SYMBOL, " ") for c in result.columns}
            )
        if mode in (RenameEnum.all, RenameEnum.index):
            result.index = [i.replace(ID_SPLIT_SYMBOL, " ") for i in result.index]
        return result

    def extract(self, experiment_data: ExperimentData):
        self._extract_by_reporters(experiment_data)


class ExperimentShell:
    def __init__(
        self,
        experiment: Experiment,
        output: Output,
        experiment_params: Optional[Dict[str, Any]] = None,
    ):
        if experiment_params:
            experiment.set_params(experiment_params)
        self._out = output
        self._experiment = experiment

    @property
    def experiment(self):
        return self._experiment

    def execute(self, data: Union[Dataset, ExperimentData]):
        if isinstance(data, Dataset):
            data = ExperimentData(data)
        result_experiment_data = self._experiment.execute(data)
        self._out.extract(result_experiment_data)
        return self._out



from typing import Optional

from ..dataset import Dataset
from .abstract import Encoder
from ..extensions.encoders import DummyEncoderExtension


class DummyEncoder(Encoder):
    @staticmethod
    def _inner_function(
        data: Dataset, target_cols: Optional[str] = None, **kwargs
    ) -> Dataset:
        if not target_cols:
            return data
        return DummyEncoderExtension().calc(
            data=data, target_cols=target_cols, **kwargs
        )



from typing import Dict, Any

from ..comparators import GroupDifference, GroupSizes, TTest, KSTest, Chi2Test
from ..dataset import (
    ExperimentData,
    Dataset,
    InfoRole,
    TreatmentRole,
    StatisticRole,
)
from ..splitters import AASplitter, AASplitterWithStratification
from ..utils import (
    ExperimentDataEnum,
    ID_SPLIT_SYMBOL,
    NotFoundInExperimentDataError,
)
from .abstract import DictReporter, Reporter


class OneAADictReporter(DictReporter):
    @staticmethod
    def rename_passed(data: Dict[str, bool]):
        return {
            c: ("NOT OK" if (v is True or v == "True") else "OK") if "pass" in c else v
            for c, v in data.items()
        }

    @staticmethod
    def _get_struct_dict(data: Dict):
        # TODO: rewrite to recursion?
        dict_result = {}
        for key, value in data.items():
            if ID_SPLIT_SYMBOL in key:
                key_split = key.split(ID_SPLIT_SYMBOL)
                if key_split[2] in ("pass", "p-value", "difference", "difference %"):
                    if key_split[0] not in dict_result:
                        dict_result[key_split[0]] = {
                            key_split[3]: {key_split[1]: {key_split[2]: value}}
                        }
                    elif key_split[3] not in dict_result[key_split[0]]:
                        dict_result[key_split[0]][key_split[3]] = {
                            key_split[1]: {key_split[2]: value}
                        }
                    elif key_split[1] not in dict_result[key_split[0]][key_split[3]]:
                        dict_result[key_split[0]][key_split[3]][key_split[1]] = {
                            key_split[2]: value
                        }
                    else:
                        dict_result[key_split[0]][key_split[3]][key_split[1]][
                            key_split[2]
                        ] = value
        return dict_result

    @staticmethod
    def _convert_struct_dict_to_dataset(data: Dict) -> Dataset:
        result = []
        for feature, groups in data.items():
            for group, tests in groups.items():
                t_values = {"feature": feature, "group": group}
                for test, values in tests.items():
                    if test == "GroupDifference":
                        t_values["difference"] = values["difference"]
                        t_values["difference %"] = (
                            values["difference %"] if "difference %" in values else None
                        )
                    else:
                        t_values[f"{test} pass"] = values["pass"]
                        t_values[f"{test} p-value"] = (
                            values["p-value"] if "p-value" in values else None
                        )
                result.append(t_values)
        result = [OneAADictReporter.rename_passed(d) for d in result]
        return Dataset.from_dict(
            result,
            roles={"feature": InfoRole(), "group": TreatmentRole()},
        )

    @staticmethod
    def convert_flat_dataset(data: Dict) -> Dataset:
        struct_dict = OneAADictReporter._get_struct_dict(data)
        return OneAADictReporter._convert_struct_dict_to_dataset(struct_dict)

    @staticmethod
    def get_splitter_id(data: ExperimentData):
        for c in [AASplitter, AASplitterWithStratification]:
            try:
                return data.get_one_id(c, ExperimentDataEnum.additional_fields)
            except NotFoundInExperimentDataError:
                pass  # The splitting was done by another class

    def extract_group_difference(self, data: ExperimentData) -> Dict[str, Any]:
        group_difference_ids = data.get_ids(GroupDifference)[GroupDifference.__name__][
            ExperimentDataEnum.analysis_tables.value
        ]
        return self._extract_from_comparators(data, group_difference_ids)

    def extract_group_sizes(self, data: ExperimentData) -> Dict[str, Any]:
        group_sizes_id = data.get_one_id(GroupSizes, ExperimentDataEnum.analysis_tables)
        return self._extract_from_comparators(data, [group_sizes_id])

    def extract_tests(self, data: ExperimentData) -> Dict[str, Any]:
        test_ids = data.get_ids(
            [TTest, KSTest, Chi2Test], searched_space=ExperimentDataEnum.analysis_tables
        )
        result = {}
        for class_, ids in test_ids.items():
            result.update(
                self._extract_from_comparators(
                    data, ids[ExperimentDataEnum.analysis_tables.value]
                )
            )
        return {k: v for k, v in result.items() if "pass" in k or "p-value" in k}

    def extract_analyzer_data(self, data: ExperimentData) -> Dict[str, Any]:
        analyzer_id = data.get_one_id(
            "OneAAStatAnalyzer", ExperimentDataEnum.analysis_tables
        )
        return self.extract_from_one_row_dataset(data.analysis_tables[analyzer_id])

    def extract_data_from_analysis_tables(self, data: ExperimentData) -> Dict[str, Any]:
        result = {}
        result.update(self.extract_group_difference(data))
        result.update(self.extract_group_sizes(data))
        result.update(self.extract_tests(data))
        result.update(self.extract_analyzer_data(data))
        if self.front:
            result = self.rename_passed(result)
        return result

    def report(self, data: ExperimentData) -> Dict[str, Any]:
        result = {
            "splitter_id": self.get_splitter_id(data),
        }
        result.update(self.extract_data_from_analysis_tables(data))
        return result


class AADatasetReporter(OneAADictReporter):
    def report(self, data: ExperimentData):
        front_buffer = self.front
        self.front = False
        dict_report = super().report(data)
        self.front = front_buffer
        return self.convert_flat_dataset(dict_report)


class AAPassedReporter(Reporter):
    @staticmethod
    def _reformat_aa_score_table(table: Dataset) -> Dataset:
        result = {}
        for ind in table.index:
            splitted_index = ind.split(ID_SPLIT_SYMBOL)
            row_index = f"{splitted_index[0]}{ID_SPLIT_SYMBOL}{splitted_index[-1]}"
            value = table.get_values(ind, "pass")
            if row_index not in result:
                result[row_index] = {splitted_index[1]: value}
            else:
                result[row_index][splitted_index[1]] = value
        result = Dataset.from_dict(result, roles={}).transpose() * 1
        return result

    @staticmethod
    def _reformat_best_split_table(table: Dataset) -> Dataset:
        passed = table.loc[:, [c for c in table.columns if (c.endswith("pass"))]]
        new_index = table.apply(
            lambda x: f"{x['feature']}{ID_SPLIT_SYMBOL}{x['group']}",
            {"index": InfoRole()},
            axis=1,
        )
        passed.index = new_index.get_values(column="index")
        passed = passed.rename(
            names={c: c[: c.rfind("pass") - 1] for c in passed.columns}
        )
        passed = passed.replace("OK", 1).replace("NOT OK", 0)
        passed = passed.astype({c: int for c in passed.columns})
        return passed

    def _detect_pass(self, analyzer_tables: Dict[str, Dataset]):
        score_table = self._reformat_aa_score_table(analyzer_tables["aa score"])
        best_split_table = self._reformat_best_split_table(
            analyzer_tables["best split statistics"]
        )
        resume_table = score_table * best_split_table
        resume_table = resume_table.apply(
            lambda x: "OK" if x.sum() > 0 else "NOT OK",
            axis=1,
            role={"result": StatisticRole()},
        )
        result = score_table.merge(
            best_split_table,
            suffixes=(" aa test", " best split"),
            left_index=True,
            right_index=True,
        )
        result = result.merge(resume_table, left_index=True, right_index=True)
        result.roles = {c: r.__class__(str) for c, r in result.roles.items()}
        result = (
            result.replace(0, "NOT OK")
            .replace(1, "OK")
            .replace("0", "NOT OK")
            .replace("1", "OK")
        )
        splitted_index = [str(i).split(ID_SPLIT_SYMBOL) for i in result.index]
        result.add_column([i[0] for i in splitted_index], role={"feature": InfoRole()})
        result.add_column([i[1] for i in splitted_index], role={"group": InfoRole()})
        result.index = range(len(splitted_index))
        return result

    def report(self, data: ExperimentData) -> Dataset:
        analyser_ids = data.get_ids(
            "AAScoreAnalyzer", ExperimentDataEnum.analysis_tables
        )
        analyser_tables = {
            id_[id_.rfind(ID_SPLIT_SYMBOL) + 1 :]: data.analysis_tables[id_]
            for id_ in analyser_ids["AAScoreAnalyzer"][
                ExperimentDataEnum.analysis_tables.value
            ]
        }
        if not analyser_tables["aa score"]:
            print("AA test cannot be performed as none of the analyzers passed")
            return None
        result = self._detect_pass(analyser_tables)
        stats_cols = ["feature", "group", "difference", "difference %"]
        differences = analyser_tables["best split statistics"].loc[
            :,
            [
                col
                for col in stats_cols
                if col in analyser_tables["best split statistics"].columns
            ],
        ]
        result = result.merge(differences, on=["feature", "group"], how="left")
        result = result[
            ["feature", "group"]
            + [c for c in result.columns if c not in ["feature", "group"]]
        ]
        return result


class AABestSplitReporter(Reporter):
    def report(self, data: ExperimentData):
        best_split_id = next(
            (c for c in data.additional_fields.columns if c.endswith("best")), []
        )
        markers = data.additional_fields.loc[:, best_split_id]
        markers = markers.rename({markers.columns[0]: "split"})
        return data.ds.merge(markers, left_index=True, right_index=True)



from typing import Optional, Literal, Union, List

from .analyzers.ab import ABAnalyzer
from .comparators import GroupDifference, GroupSizes, TTest, UTest, Chi2Test
from .dataset import TargetRole, TreatmentRole
from .experiments.base import Experiment, OnRoleExperiment
from .ui.ab import ABOutput
from .ui.base import ExperimentShell
from .utils import ABNTestMethodsEnum


class ABTest(ExperimentShell):

    @staticmethod
    def _make_experiment(additional_tests, multitest_method):
        test_mapping = {
            "t-test": TTest(compare_by="groups", grouping_role=TreatmentRole()),
            "u-test": UTest(compare_by="groups", grouping_role=TreatmentRole()),
            "chi2-test": Chi2Test(compare_by="groups", grouping_role=TreatmentRole()),
        }
        on_role_executors = [GroupDifference(grouping_role=TreatmentRole())]
        additional_tests = ["t-test"] if additional_tests is None else additional_tests
        additional_tests = (
            additional_tests
            if isinstance(additional_tests, List)
            else [additional_tests]
        )
        for i in additional_tests:
            on_role_executors += [test_mapping[i]]
        return Experiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole()),
                OnRoleExperiment(
                    executors=on_role_executors,
                    role=TargetRole(),
                ),
                ABAnalyzer(
                    multitest_method=(
                        ABNTestMethodsEnum(multitest_method)
                        if multitest_method
                        else None
                    )
                ),
            ]
        )

    def __init__(
        self,
        additional_tests: Union[
            Literal["t-test", "u-test", "chi2-test"],
            List[Literal["t-test", "u-test", "chi2-test"]],
            None,
        ] = None,
        multitest_method: Optional[
            Literal[
                "bonferroni",
                "sidak",
                "holm-sidak",
                "holm",
                "simes-hochberg",
                "hommel",
                "fdr_bh",
                "fdr_by",
                "fdr_tsbh",
                "fdr_tsbhy",
                "quantile",
            ]
        ] = "holm",
    ):
        super().__init__(
            experiment=self._make_experiment(additional_tests, multitest_method),
            output=ABOutput(),
        )



from typing import Union

from .base import Output
from ..analyzers.ab import ABAnalyzer
from ..comparators import GroupSizes, GroupDifference
from ..dataset import ExperimentData, Dataset, TreatmentRole, StatisticRole
from ..reporters.ab import ABDatasetReporter
from ..utils import ExperimentDataEnum, ID_SPLIT_SYMBOL


class ABOutput(Output):
    multitest: Union[Dataset, str]

    def __init__(self):
        self._groups = []
        super().__init__(resume_reporter=ABDatasetReporter())

    def _extract_multitest_result(self, experiment_data: ExperimentData):
        multitest_id = experiment_data.get_one_id(
            ABAnalyzer, ExperimentDataEnum.analysis_tables
        )
        if multitest_id and "MultiTest" in multitest_id:
            self.multitest = experiment_data.analysis_tables[multitest_id]
        else:
            self.multitest = (
                "There was less than three groups or multitest method wasn't provided"
            )

    def _extract_differences(self, experiment_data: ExperimentData):
        targets = []
        groups = []
        ids = experiment_data.get_ids(
            GroupDifference,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupDifference"]["analysis_tables"]
        self._groups = list(
            experiment_data.groups[
                experiment_data.ds.search_columns(TreatmentRole())[0]
            ].keys()
        )[1:]
        for i in self._groups:
            groups += [i] * len(ids)
        diff = Dataset.create_empty()
        for i in range(len(ids)):
            diff = diff.append(experiment_data.analysis_tables[ids[i]])
            targets += [ids[i].split(ID_SPLIT_SYMBOL)[-1]]
        return diff.add_column(groups, role={"group": StatisticRole()}).add_column(
            targets * len(self._groups), role={"feature": StatisticRole()}
        )

    def _extract_sizes(self, experiment_data: ExperimentData):
        ids = experiment_data.get_ids(
            GroupSizes,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupSizes"]["analysis_tables"]
        self.sizes = experiment_data.analysis_tables[ids[0]].add_column(
            self._groups, role={"group": StatisticRole()}
        )

    def extract(self, experiment_data: ExperimentData):
        super().extract(experiment_data)
        self._extract_differences(experiment_data)
        self._extract_multitest_result(experiment_data)
        self._extract_sizes(experiment_data)



from typing import Optional, List

from scipy.stats import ttest_ind, ks_2samp, mannwhitneyu  # type: ignore

from .abstract import StatHypothesisTesting
from ..dataset import Dataset
from ..extensions.scipy_stats import (
    TTestExtension,
    KSTestExtension,
    UTestExtension,
    Chi2TestExtension,
)
from ..utils.constants import NUMBER_TYPES_LIST


class TTest(StatHypothesisTesting):
    @property
    def search_types(self) -> Optional[List[type]]:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Dataset:
        return TTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class KSTest(StatHypothesisTesting):

    @property
    def search_types(self) -> Optional[List[type]]:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Dataset:
        return KSTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class UTest(StatHypothesisTesting):

    @property
    def search_types(self) -> Optional[List[type]]:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Dataset:
        return UTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class Chi2Test(StatHypothesisTesting):

    @property
    def search_types(self) -> Optional[List[type]]:
        return [str]

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Dataset:
        return Chi2TestExtension(reliability=kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )



from .abstract import Reporter, DictReporter, DatasetReporter
from .homo import HomoDictReporter, HomoDatasetReporter

__all__ = [
    "Reporter",
    "DictReporter",
    "DatasetReporter",
    "HomoDictReporter",
    "HomoDatasetReporter",
]



from .abstract import StatHypothesisTesting, Comparator
from .comparators import GroupDifference, GroupSizes
from .distances import MahalanobisDistance
from .hypothesis_testing import TTest, KSTest, UTest, Chi2Test
from .test_power import TestPower, MDEBySize

__all__ = [
    "GroupDifference",
    "GroupSizes",
    "StatHypothesisTesting",
    "Comparator",
    "KSTest",
    "UTest",
    "TTest",
    "Chi2Test",
    "MahalanobisDistance",
    "TestPower",
    "MDEBySize",
]



from abc import ABC
from typing import Optional, Dict, Union
from copy import deepcopy

from ..utils import (
    TargetRoleTypes,
    DefaultRoleTypes,
    CategoricalTypes,
    RoleNameType,
)


class ABCRole(ABC):
    """Base abstract class for all roles.
    Base class for defining different roles within a data set. Each role corresponds to a specific
    type of data handling or processing functionality. This class is abstract and should be
    subclassed to implement specific roles.
    Attributes:
        _role_name (RoleNameType): The default name of the role, intended to be overridden in subclasses.
        data_type (Optional[DefaultRoleTypes]): The type of data associated with this role,
                                                which can be any defined type or None. This
                                                specifies how data should be treated or processed
                                                within the given role context.
    """

    _role_name: RoleNameType = "Abstract"

    def __init__(self, data_type: Optional[DefaultRoleTypes] = None):
        """Initializes a new instance of ABCRole with an optional data type.
        Attributes:
            data_type (Optional[DefaultRoleTypes], optional): Specifies the data type associated with the role.
                                                              This can influence how data is validated, processed,
                                                              or interpreted. Defaults to None.
        """
        self.data_type = data_type

    @property
    def role_name(self) -> str:
        """Retrieves the name of the role.
        Returns:
            RoleNameType: A string that represents the name of the role. This is typically a descriptive
                 label that identifies the role's purpose or function in data handling.
        """
        return self._role_name

    def __repr__(self) -> str:
        """Provides a string representation of the role, which includes its name and data type.
        Returns:
            str: A string representation of the role, combining the role's name with its data type,
                 formatted for clarity and easy understanding. For example, "Abstract(None)" if
                 no data type is specified.
        """
        return f"{self._role_name}({self.data_type})"


class InfoRole(ABCRole):
    """Role for information column.
    Represents informational data that doesn't directly affect the analysis but may be useful
    for reporting, debugging, or tracking purposes.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Info'.
    """

    _role_name: RoleNameType = "Info"


class StratificationRole(ABCRole):
    """Role for column that need to be stratified.
    Used to ensure that different treatment groups are statistically comparable by balancing
    them based on the stratification variables. This helps in reducing bias in the treatment
    effect estimation by equalizing covariates that could influence the outcome.
    Attributes:
        data_type (Optional[CategoricalTypes]): Specifies the data type for the stratification feature.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Stratification'.
    """

    _role_name: RoleNameType = "Stratification"

    def __init__(self, data_type: Optional[CategoricalTypes] = None):
        super().__init__(data_type)


class GroupingRole(ABCRole):
    """Role for column that need to be grouped by categorical variables.
    Facilitates comparisons within homogenous subgroups. This role is used to ensure that
    analyses or comparisons are made within defined categories to maintain validity and relevance,
    particularly when examining effects that may vary by group.
    Attributes:
        data_type (Optional[CategoricalTypes]): The type of the categorical data used for grouping.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Grouping'.
    """

    _role_name: RoleNameType = "Grouping"

    def __init__(self, data_type: Optional[CategoricalTypes] = None):
        super().__init__(data_type)


class TreatmentRole(ABCRole):
    """Role for column that shows treatment.
    Designates the treatment assignment of subjects in an experimental study. This role is critical
    for identifying which group a subject belongs to, facilitating the assessment of the treatment
    effects across different conditions.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Treatment'.
    """

    _role_name: RoleNameType = "Treatment"


class TargetRole(ABCRole):
    """Role for column that shows target.
    Represents the primary outcome variable of interest in an analysis or study. Target roles
    are crucial for defining what the analysis aims to predict or explain.
    Attributes:
        data_type (Optional[TargetRoleTypes]): The type of the target data, which influences
                                               how the data is processed or analyzed.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Target'.
    """

    _role_name: RoleNameType = "Target"

    def __init__(self, data_type: Optional[TargetRoleTypes] = None):
        super().__init__(data_type)


class FeatureRole(ABCRole):
    """Roles for column that shows feature.
    Used for features that directly influence the outcome of the analysis. These are key variables
    that algorithms use to learn patterns or make decisions.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Feature'.
    """

    _role_name: RoleNameType = "Feature"


class PreTargetRole(ABCRole):
    """Roles for column that shows Pre target.
    Represents the value of the target variable at a previous time point. This is often required
    for algorithms that need to consider historical outcomes to make predictions or assessments.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'PreTarget'.
    """

    _role_name: RoleNameType = "PreTarget"

    def __init__(self, data_type: Optional[TargetRoleTypes] = None):
        super().__init__(data_type)


class StatisticRole(ABCRole):
    """Role for statistic column.
    Designated for columns that are used to compute statistical parameters or metrics.
    This role facilitates the identification and processing of data that contributes to
    statistical analysis, ensuring that relevant calculations are performed accurately.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Statistic'.
    """

    _role_name: RoleNameType = "Statistic"


class ResumeRole(ABCRole):
    _role_name = "Resume"


class FilterRole(ABCRole):
    _role_name: RoleNameType = "Filter"


class ConstGroupRole(ABCRole):
    _role_name: RoleNameType = "ConstGroup"


# ___________________________________________________________________________________________
class TempRole(ABCRole):
    _role_name: RoleNameType = "Temp"


class TempTreatmentRole(TempRole, TreatmentRole):
    """Role for temp treatment column.
    A temporary role used for transient treatment assignments in executables or during
    experimental processing phases. This role can help manage dynamic treatment group
    assignments that are not final but necessary for intermediate calculations.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'TempTreatment'.
    """

    _role_name: RoleNameType = "TempTreatment"


class TempTargetRole(TempRole, TargetRole):
    """Role for temp target column.
    A temporary role for target variables used in intermediate stages of data processing
    or during certain executions in the analysis pipeline. It facilitates operations on
    target data that may not be final or are used for testing scenarios.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'TempTarget'.
    """

    _role_name: RoleNameType = "TempTarget"


class TempGroupingRole(TempRole, GroupingRole):
    """Role for temp grouping column.
    Used temporarily to manage groups in the context of execution cursors or during
    specific phases of data manipulation. This role assists in keeping track of groups
    that are in flux or under evaluation in analytical processes.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'TempGrouping'.
    """

    _role_name: RoleNameType = "TempGrouping"


class DefaultRole(ABCRole):
    """The default role for a  newly created columns or any column for which another role has not been defined.
    Attributes:
        _role_name (RoleNameType): A name identifying the role as 'Default'.
    """

    _role_name: RoleNameType = "Default"


class ReportRole(ABCRole):
    """ """

    _role_name: RoleNameType = "Report"


# ___________________________________________________________________________________________
class AdditionalRole(ABCRole):
    """ """

    _role_name: RoleNameType = "Additional"


class AdditionalTreatmentRole(AdditionalRole):
    """ """

    _role_name: RoleNameType = "AdditionalTreatment"


class AdditionalGroupingRole(AdditionalRole):
    """ """

    _role_name: RoleNameType = "AdditionalGrouping"


class AdditionalTargetRole(AdditionalRole):
    """ """

    _role_name: RoleNameType = "AdditionalTarget"


class AdditionalPreTargetRole(AdditionalRole):
    """ """

    _role_name: RoleNameType = "AdditionalPreTarget"


class AdditionalMatchingRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalMatching"


default_roles: Dict[RoleNameType, ABCRole] = {
    "info": InfoRole(),
    "default": DefaultRole(),
    "feature": FeatureRole(),
    "treatment": TreatmentRole(),
    "grouping": GroupingRole(),
    "target": TargetRole(),
    "pretarget": PreTargetRole(),
    "stratification": StratificationRole(),
    "statistic": StatisticRole(),
    "filter": FilterRole(),
    "additionaltreatment": AdditionalTreatmentRole(),
    "additionalgrouping": AdditionalGroupingRole(),
    "additionaltarget": AdditionalTargetRole(),
    "additionalpretarget": AdditionalPreTargetRole(),
}



from copy import deepcopy
from typing import Optional, Any, Union, List

from ..comparators import TTest, UTest
from ..dataset import (
    Dataset,
    ExperimentData,
    StatisticRole,
    TreatmentRole,
    TargetRole,
)
from ..experiments.base import Executor
from ..extensions.statsmodels import MultiTest, MultitestQuantile
from ..utils import (
    ID_SPLIT_SYMBOL,
    NAME_BORDER_SYMBOL,
    BackendsEnum,
    ExperimentDataEnum,
    ABNTestMethodsEnum,
)


class ABAnalyzer(Executor):
    def __init__(
        self,
        multitest_method: Optional[ABNTestMethodsEnum] = None,
        alpha: float = 0.05,
        equal_variance: bool = True,
        quantiles: Optional[Union[float, List[float]]] = None,
        iteration_size: int = 20000,
        random_state: Optional[int] = None,
        key: Any = "",
    ):
        self.multitest_method = multitest_method
        self.alpha = alpha
        self.equal_variance = equal_variance
        self.quantiles = quantiles
        self.iteration_size = iteration_size
        self.random_state = random_state
        super().__init__(key)

    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.analysis_tables,
            self.id + key if key else self.id,
            value,
        )

    def execute_multitest(self, data: ExperimentData, p_values: Dataset, **kwargs):
        group_field = data.ds.search_columns(TreatmentRole())[0]
        target_fields = data.ds.search_columns(TargetRole(), search_types=[int, float])
        if self.multitest_method and len(data.groups[group_field]) > 2:
            if self.multitest_method != ABNTestMethodsEnum.quantile:
                multitest_result = MultiTest(self.multitest_method).calc(
                    p_values, **kwargs
                )
                groups = []
                for i in list(data.groups[group_field].keys())[1:]:
                    groups += [i] * len(target_fields)
                multitest_result = multitest_result.add_column(
                    groups
                    * (
                        len(multitest_result)
                        // len(target_fields)
                        // (len(data.groups[group_field]) - 1)
                    ),
                    role={"group": StatisticRole()},
                )

            else:
                multitest_result = Dataset.create_empty()
                for target_field in target_fields:
                    multitest_result = multitest_result.append(
                        MultitestQuantile(
                            self.alpha,
                            self.iteration_size,
                            self.equal_variance,
                            self.random_state,
                        ).calc(
                            p_values,
                            group_field=group_field,
                            target_field=target_field,
                            quantiles=self.quantiles,
                        )
                    )
            return self._set_value(data, multitest_result, key="MultiTest")
        return data

    def _add_pvalues(self, multitest_pvalues, value, field):
        if (
            self.multitest_method
            and field == "p-value"
            and self.multitest_method != "quantile"
        ):
            multitest_pvalues = multitest_pvalues.append(value)
        return multitest_pvalues

    def execute(self, data: ExperimentData) -> ExperimentData:
        executor_ids = data.get_ids([TTest, UTest])
        num_groups = len(data.groups[data.ds.search_columns(TreatmentRole())[0]]) - 1
        groups = list(data.groups[data.ds.search_columns(TreatmentRole())[0]].items())
        multitest_pvalues = Dataset.create_empty()
        analysis_data = {}
        for c, spaces in executor_ids.items():
            analysis_ids = spaces.get("analysis_tables", [])
            if len(analysis_ids) == 0:
                continue
            t_data = deepcopy(data.analysis_tables[analysis_ids[0]])
            for aid in analysis_ids[1:]:
                t_data = t_data.append(data.analysis_tables[aid])
            if len(analysis_ids) < len(t_data):
                analysis_ids *= num_groups
            t_data.data.index = analysis_ids
            for f in ["p-value", "pass"]:
                for i in range(0, len(analysis_ids), len(analysis_ids) // num_groups):
                    value = t_data.iloc[i : i + len(analysis_ids) // num_groups][f]
                    multitest_pvalues = self._add_pvalues(multitest_pvalues, value, f)
                    analysis_data[f"{c} {f} {groups[i//num_groups + 1][0]}"] = (
                        value.mean()
                    )
            if c not in ["UTest", "TTest"]:
                indexes = t_data.index
                values = t_data.data.values.tolist()
                for idx, value in zip(indexes, values):
                    name = idx.split(ID_SPLIT_SYMBOL)[-1]
                    analysis_data[
                        f"{c} {name[name.find(NAME_BORDER_SYMBOL) + 1: name.rfind(NAME_BORDER_SYMBOL)]}"
                    ] = value[0]

        analysis_dataset = Dataset.from_dict(
            [analysis_data],
            {f: StatisticRole() for f in analysis_data},
            BackendsEnum.pandas,
        )
        data = self.execute_multitest(
            data,
            (
                multitest_pvalues
                if not multitest_pvalues.is_empty()
                and self.multitest_method != ABNTestMethodsEnum.quantile
                else data.ds
            ),
        )

        return self._set_value(data, analysis_dataset)



import sys
from pathlib import Path
from typing import Iterable, Union

import numpy as np
import pandas as pd

ROOT = Path("").absolute().parents[0]
sys.path.append(str(ROOT))


def set_nans(
    data: pd.DataFrame,
    na_step: Union[Iterable[int], int] = None,
    nan_cols: Union[Iterable[str], str] = None,
):
    """Fill some values with NaN/

    Args:
        data: input dataframe
        na_step:
            num or list of nums of period to make NaN (step of range)
            If list - iterates accordingly order of columns
        nan_cols:
            name of one or several columns to fill with NaN
            If list - iterates accordingly order of na_step

    Returns:
        data: dataframe with some NaNs
    """
    if (nan_cols is not None) or (na_step is not None):
        # correct type of columns to iterate

        #  number of nans
        if na_step is None:
            na_step = [10]
            print(f"No na_step specified: set to {na_step}")
        elif not isinstance(na_step, Iterable):
            na_step = [na_step]

        #  columns
        if nan_cols is None:
            nan_cols = list(data.columns)
            print("No nan_cols specified. Setting NaNs applied to all columns")
        elif not isinstance(nan_cols, Iterable):
            nan_cols = [nan_cols]

        # correct length of two lists
        if len(na_step) > len(nan_cols):
            na_step = na_step[: len(nan_cols)]
            # print('Length of na_step is bigger than length of columns. Used only first values') TODO: set to logging
        elif len(na_step) < len(nan_cols):
            na_step = na_step + [na_step[-1]] * (len(nan_cols) - len(na_step))
            # print('Length of na_step is less than length of columns. Used last value several times')

        # create list of indexes to fill with na
        nans_indexes = [
            list(range(i, len(data), period)) for i, period in enumerate(na_step)
        ]

        for i in range(len(nan_cols)):
            try:
                data.loc[nans_indexes[i], nan_cols[i]] = np.nan
            except KeyError:
                print(
                    f"There is no column {nan_cols[i]} in data. No nans in this column will be added."
                )
    else:
        print("No NaN added")

    return data


def create_test_data(
    num_users: int = 10000,
    na_step: Union[Iterable[int], int] = None,
    nan_cols: Union[Iterable[str], str] = None,
    file_name: str = None,
    exact_ATT: int = 100,
    rs=None,
):
    """Creates data for tutorial.

    Args:
        num_users: num of strings
        na_step:
            num or list of nums of period to make NaN (step of range)
            If list - iterates accordingly order of columns
        nan_cols:
            name of one or several columns to fill with NaN
            If list - iterates accordingly order of na_step
        file_name: name of file to save; doesn't save file if None
        exact_ATT: an accurate synthetic effect

    Returns:
        data: dataframe with
    """
    if rs is not None:
        np.random.seed(rs)

    if (nan_cols is not None) and isinstance(nan_cols, str):
        nan_cols = [nan_cols]
    # Simulating dataset with known effect size
    num_months = 12

    # signup_months == 0 means customer did not sign up
    signup_months = np.random.choice(
        np.arange(1, num_months), num_users
    ) * np.random.randint(0, 2, size=num_users)

    data = pd.DataFrame(
        {
            "user_id": np.repeat(np.arange(num_users), num_months),
            "signup_month": np.repeat(
                signup_months, num_months
            ),  # signup month == 0 means customer did not sign up
            "month": np.tile(
                np.arange(1, num_months + 1), num_users
            ),  # months are from 1 to 12
            "spend": np.random.poisson(500, num_users * num_months),
        }
    )

    # A customer is in the treatment group if and only if they signed up
    data["treat"] = data["signup_month"] > 0

    # Simulating an effect of month (monotonically decreasing--customers buy less later in the year)
    data["spend"] = data["spend"] - data["month"] * 10

    # Simulating a simple treatment effect of 100
    after_signup = (data["signup_month"] < data["month"]) & (data["treat"])
    data.loc[after_signup, "spend"] = data[after_signup]["spend"] + exact_ATT

    # Setting the signup month (for ease of analysis)
    i = 3
    data = (
        data.groupby(["user_id", "signup_month", "treat"])
        .apply(
            lambda x: pd.Series(
                {
                    "pre_spends": x.loc[x.month < i, "spend"].mean(),
                    "post_spends": x.loc[x.month > i, "spend"].mean(),
                }
            )
        )
        .reset_index()
    )

    # Additional category features
    gender_i = np.random.choice(a=[0, 1], size=data.user_id.nunique())
    gender = [["M", "F"][i] for i in gender_i]

    age = np.random.choice(a=range(18, 70), size=data.user_id.nunique())

    industry_i = np.random.choice(a=range(1, 3), size=data.user_id.nunique())
    industry_names = ["Finance", "E-commerce", "Logistics"]
    industry = [industry_names[i] for i in industry_i]

    data["age"] = age
    data["gender"] = gender
    data["industry"] = industry
    data["industry"] = data["industry"].astype("str")
    data["treat"] = data["treat"].astype(int)

    # input nans in data if needed
    data = set_nans(data, na_step, nan_cols)

    if file_name is not None:
        data.to_csv(ROOT / f"{file_name}.csv", index=False)

    return data


def sigmoid(x: np.ndarray) -> np.ndarray:
    """Logistic sigmoid ufunc for ndarrays.

    The sigmoid function, also known as the logistic sigmoid function,
    is defined as sigmoid(x) = $\fraq{1}{(1+\exp{-x})}$.
    It is the inverse of the logit function.

    Args:
        x:
            Input array.

    Returns:
        Sigmoid function of x.
    """
    return 1 / (1 + np.exp(-x))


def sigmoid_division(x, dependent_division=True) -> np.ndarray:
    """
    Division method via sigmoid function.

    Args:
        x:
            Input array
        dependent_division:
            If True - returns the binary vector dependent on the X.
            If False - returns the binary vector independent on the X.

    Returns:
        Binary array coordinated with variable x.
    """
    if dependent_division:
        division = np.random.binomial(1, sigmoid((x - x.mean()) / x.std()))
    else:
        division = np.random.binomial(1, 0.5, size=len(x))
    return division


def gen_special_medicine_df(
    data_size=100, *, dependent_division=True, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises dependent/independent group splitting.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about patient treatment.
    """
    if random_state is not None:
        np.random.seed(random_state)

    disease_degree = np.random.choice(
        [1, 2, 3, 4, 5], p=[0.3, 0.3, 0.2, 0.1, 0.1], size=data_size
    ).astype(int)

    age = np.random.normal(40, scale=8, size=data_size).astype(int)

    age_effect = (age ^ 2 - 400) / 1000

    experimental_treatment = sigmoid_division(disease_degree, dependent_division)

    residual_lifetime = np.random.exponential(
        17 - 2.5 * disease_degree + 1 * experimental_treatment - age_effect
    )

    df = pd.DataFrame(
        dict(
            age=age,
            disease_degree=disease_degree,
            experimental_treatment=experimental_treatment,
            residual_lifetime=residual_lifetime,
        )
    )
    return df


def gen_oracle_df(
    data_size=8, *, dependent_division=True, factual_only=False, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises factual and counterfactual outcomes.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        factual_only:
            Defines the display of counterfactual states.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about
        factual and counterfactual states of the target.
    """
    if random_state is not None:
        np.random.seed(random_state)

    treatment = np.random.binomial(1, 0.5, size=data_size)

    if dependent_division:
        target_feature = np.random.binomial(1, 0.3 + 0.4 * treatment)
    else:
        target_feature = np.random.binomial(1, 0.5, size=data_size)

    target_untreated = (
        np.random.uniform(low=300, high=800, size=data_size).round(-2).astype(int)
    )

    target_treated = target_untreated + 50 + target_feature * 100

    if factual_only:
        target_untreated = np.where(1 - treatment, target_untreated, np.nan)
        target_treated = np.where(treatment, target_treated, np.nan)

    y_factual = np.where(treatment, target_treated, target_untreated).astype("int")

    treatment_effect = target_treated - target_untreated

    df = pd.DataFrame(
        dict(
            X=target_feature,
            Target_untreated=target_untreated,
            Target_treated=target_treated,
            Treatment=treatment,
            Target=y_factual,
            TE=treatment_effect,
        )
    )
    return df


def gen_control_variates_df(
    data_size=1000, *, dependent_division=True, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises 0-variation outcome mixed with linear x dependency.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about
        features and lagged features.
    """
    if random_state is not None:
        np.random.seed(random_state)

    mean_x_feature = np.random.uniform(0, 5, size=data_size)

    x_feature_lag_1 = np.random.normal(mean_x_feature, 2)
    x_feature = np.random.normal(mean_x_feature, 2)

    treatment = sigmoid_division(x_feature_lag_1, dependent_division)

    target_lag_1 = 200 + x_feature_lag_1 * 100
    target_factual = 200 + x_feature * 100 + treatment * 10

    df = pd.DataFrame(
        dict(
            X_lag_1=x_feature_lag_1,
            Target_lag_1=target_lag_1,
            X=x_feature,
            Treatment=treatment,
            Target=target_factual,
        )
    )
    return df



import warnings
from typing import List, Literal, Union

from .experiments import GroupExperiment
from .reporters.matching import MatchingDatasetReporter
from .analyzers.matching import MatchingAnalyzer
from .comparators import TTest
from .comparators.distances import MahalanobisDistance
from .dataset import TreatmentRole, TargetRole, AdditionalTargetRole
from .executor import Executor
from .experiments.base import Experiment
from .ml.faiss import FaissNearestNeighbors
from .operators.operators import MatchingMetrics, Bias
from .ui.base import ExperimentShell
from .ui.matching import MatchingOutput


class Matching(ExperimentShell):

    @staticmethod
    def _make_experiment(
        group_match: bool = False,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        metric: Literal["atc", "att", "ate"] = "ate",
        bias_estimation: bool = True,
        quality_tests: Union[
            Literal["smd", "psi", "ks-test", "repeats", "auto"],
            List[Literal["smd", "psi", "ks-test", "repeats", "auto"]],
        ] = "auto",
    ) -> Experiment:
        distance_mapping = {
            "mahalanobis": MahalanobisDistance(grouping_role=TreatmentRole())
        }
        two_sides = metric == "ate"
        test_pairs = metric == "atc"
        executors: List[Executor] = [
            FaissNearestNeighbors(
                grouping_role=TreatmentRole(),
                two_sides=two_sides,
                test_pairs=test_pairs,
            )
        ]
        if bias_estimation:
            executors += [
                Bias(grouping_role=TreatmentRole(), target_roles=[TargetRole()]),
            ]
        executors += [
            MatchingMetrics(
                grouping_role=TreatmentRole(),
                target_roles=[TargetRole()],
                metric=metric,
            ),
            TTest(
                compare_by="columns_in_groups",
                baseline_role=AdditionalTargetRole(),
                target_role=TargetRole(),
                grouping_role=TreatmentRole(),
            ),
            MatchingAnalyzer(),
        ]
        if quality_tests != "auto":
            warnings.warn("Now quality tests aren't supported yet")
        return (
            Experiment(
                executors=(
                    executors
                    if distance == "l2"
                    else [distance_mapping[distance]] + executors
                )
            )
            if not group_match
            else GroupExperiment(
                executors=(
                    executors
                    if distance == "l2"
                    else [distance_mapping[distance]] + executors
                ),
                reporter=MatchingDatasetReporter(),
            )
        )

    def __init__(
        self,
        group_match: bool = False,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        metric: Literal["atc", "att", "ate"] = "ate",
        bias_estimation: bool = True,
        quality_tests: Union[
            Literal["smd", "psi", "ks-test", "repeats", "auto"],
            List[Literal["smd", "psi", "ks-test", "repeats", "auto"]],
        ] = "auto",
    ):
        super().__init__(
            experiment=self._make_experiment(
                group_match, distance, metric, bias_estimation, quality_tests
            ),
            output=MatchingOutput(GroupExperiment if group_match else MatchingAnalyzer),
        )



from abc import ABC, abstractmethod
from typing import Dict, Any, Union, List

from ..dataset import ExperimentData, Dataset
from ..dataset.roles import ReportRole
from ..utils import ID_SPLIT_SYMBOL
from ..utils.errors import AbstractMethodError


class Reporter(ABC):
    @abstractmethod
    def report(self, data: ExperimentData):
        raise AbstractMethodError


class DictReporter(Reporter, ABC):
    def __init__(self, front=True):
        self.front = front

    @staticmethod
    def extract_from_one_row_dataset(data: Dataset) -> Dict[str, Any]:
        return {k: v[0] for k, v in data.to_dict()["data"]["data"].items()}

    def _extract_from_comparator(self, data: ExperimentData, comparator_id: str):
        result = {}
        field = comparator_id[comparator_id.rfind(ID_SPLIT_SYMBOL) + 1 :]
        executor_name = comparator_id[: comparator_id.find(ID_SPLIT_SYMBOL)]
        sep = " " if self.front else ID_SPLIT_SYMBOL
        analysis_dict = data.analysis_tables[comparator_id].to_dict()["data"]
        for i, index_value in enumerate(analysis_dict["index"]):
            for k, v in analysis_dict["data"].items():
                key = sep.join(
                    [field, executor_name, k, str(index_value)]
                    if field
                    else [executor_name, k, str(index_value)]
                )
                result[key] = v[i]
        return result

    def _extract_from_comparators(
        self, data: ExperimentData, comparator_ids: List[str]
    ) -> Dict[str, Any]:
        result = {}
        for comparator_id in comparator_ids:
            result.update(self._extract_from_comparator(data, comparator_id))
        return result

    @abstractmethod
    def report(self, data: ExperimentData) -> Dict:
        raise AbstractMethodError


class OnDictReporter(Reporter, ABC):
    def __init__(self, dict_reporter: DictReporter) -> None:
        self.dict_reporter = dict_reporter


class DatasetReporter(OnDictReporter):
    def report(self, data: ExperimentData) -> Union[Dict[str, Dataset], Dataset]:
        dict_result = self.dict_reporter.report(data)
        return self.convert_to_dataset(
            dict_result
        )  #   TODO: change to DatasetAdapter.to_dataset()

    @staticmethod
    def convert_to_dataset(data: Dict) -> Union[Dict[str, Dataset], Dataset]:
        return Dataset.from_dict(roles={k: ReportRole() for k in data}, data=[data])



from typing import Optional, Dict, Any, Iterable

from .forks.aa import IfAAExecutor
from .analyzers.aa import OneAAStatAnalyzer, AAScoreAnalyzer
from .comparators import GroupDifference, GroupSizes
from .comparators.abstract import Comparator
from .comparators.hypothesis_testing import TTest, KSTest, Chi2Test
from .dataset import AdditionalTreatmentRole
from .dataset import TargetRole
from .experiments.base import Experiment, OnRoleExperiment
from .experiments.base_complex import ParamsExperiment, IfParamsExperiment
from .reporters import DatasetReporter
from .reporters.aa import OneAADictReporter
from .splitters import AASplitter, AASplitterWithStratification
from .ui.aa import AAOutput
from .ui.base import ExperimentShell
from .utils import SpaceEnum

AA_METRICS = Experiment(
    executors=[
        GroupSizes(grouping_role=AdditionalTreatmentRole()),
        OnRoleExperiment(
            executors=[
                GroupDifference(
                    compare_by="groups", grouping_role=AdditionalTreatmentRole()
                ),
                TTest(compare_by="groups", grouping_role=AdditionalTreatmentRole()),
                KSTest(compare_by="groups", grouping_role=AdditionalTreatmentRole()),
                Chi2Test(compare_by="groups", grouping_role=AdditionalTreatmentRole()),
            ],
            role=TargetRole(),
        ),
        OneAAStatAnalyzer(),
    ]
)

ONE_AA_TEST = Experiment(executors=[AASplitter(), AA_METRICS])
ONE_AA_TEST_WITH_STRATIFICATION = Experiment(
    executors=[AASplitterWithStratification(), AA_METRICS]
)

AA_TEST = Experiment(
    [
        ParamsExperiment(
            executors=([ONE_AA_TEST]),
            params={
                AASplitter: {"random_state": range(2000), "control_size": [0.5]},
                Comparator: {
                    "grouping_role": [AdditionalTreatmentRole()],
                    "space": [SpaceEnum.additional],
                },
            },
            reporter=DatasetReporter(OneAADictReporter(front=False)),
        ),
        AAScoreAnalyzer(),
    ],
    key="AATest",
)
AA_TEST_WITH_STRATIFICATION = Experiment(
    [
        ParamsExperiment(
            executors=([ONE_AA_TEST_WITH_STRATIFICATION]),
            params={
                AASplitter: {"random_state": range(2000), "control_size": [0.5]},
                Comparator: {
                    "grouping_role": [AdditionalTreatmentRole()],
                    "space": [SpaceEnum.additional],
                },
            },
            reporter=DatasetReporter(OneAADictReporter(front=False)),
        ),
        AAScoreAnalyzer(),
    ],
    key="AATest",
)


class AATest(ExperimentShell):
    @staticmethod
    def _prepare_params(
        n_iterations: int,
        control_size: float,
        random_states: Optional[Iterable[int]] = None,
        sample_size: Optional[float] = None,
        additional_params: Optional[Dict[str, Any]] = None,
    ) -> Dict[type, Dict[str, Any]]:
        random_states = random_states or range(n_iterations)
        additional_params = additional_params or {}
        params = {
            AASplitter: {
                "random_state": random_states,
                "control_size": [control_size],
                "sample_size": [sample_size],
            },
            Comparator: {
                "grouping_role": [AdditionalTreatmentRole()],
                "space": [SpaceEnum.additional],
            },
        }

        params.update(additional_params)
        return params

    def __init__(
        self,
        n_iterations: int = 2000,
        control_size: float = 0.5,
        stratification: bool = False,
        sample_size: Optional[float] = None,
        additional_params: Optional[Dict[str, Any]] = None,
        random_states: Optional[Iterable[int]] = None,
    ):

        super().__init__(
            experiment=Experiment(
                [
                    ParamsExperiment(
                        executors=(
                            [
                                (
                                    ONE_AA_TEST_WITH_STRATIFICATION
                                    if stratification
                                    else ONE_AA_TEST
                                )
                            ]
                        ),
                        params=self._prepare_params(
                            n_iterations,
                            control_size,
                            random_states,
                            sample_size,
                            additional_params,
                        ),
                        reporter=DatasetReporter(OneAADictReporter(front=False)),
                    ),
                    IfParamsExperiment(
                        executors=(
                            [
                                (
                                    ONE_AA_TEST_WITH_STRATIFICATION
                                    if stratification
                                    else ONE_AA_TEST
                                )
                            ]
                        ),
                        params=self._prepare_params(
                            n_iterations,
                            control_size,
                            random_states,
                            additional_params,
                        ),
                        reporter=DatasetReporter(OneAADictReporter(front=False)),
                        stopping_criterion=IfAAExecutor(sample_size=sample_size),
                    ),
                    AAScoreAnalyzer(),
                ],
                key="AATest",
            ),
            output=AAOutput(),
        )



from abc import ABC, abstractmethod
from typing import Any, Optional

from .comparators import Comparator
from ..dataset import ExperimentData, ABCRole, Dataset
from ..utils import SpaceEnum, ExperimentDataEnum


class TestPower(Comparator, ABC):
    def __init__(
        self,
        grouping_role: Optional[ABCRole] = None,
        # space: SpaceEnum = SpaceEnum.auto,
        significance: float = 0.95,
        power: float = 0.8,
        key: Any = "",
    ):
        super().__init__(
            compare_by="groups",
            grouping_role=grouping_role,
            # space=space,
            key=key,
        )
        self.significance = significance
        self.power = power

    @classmethod
    @abstractmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        significance: float = 0.95,
        power: float = 0.8,
        **kwargs
    ) -> float:
        pass

    def execute(self, data: ExperimentData) -> ExperimentData:
        return super().execute(data)


class MDEBySize(TestPower):
    def _set_value(
        self, data: ExperimentData, value: Optional[Dataset] = None, key: Any = None
    ) -> ExperimentData:
        data.set_value(
            ExperimentDataEnum.variables,
            self.id,
            value,
        )
        return data

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        significance: float = 0.95,
        power: float = 0.8,
        **kwargs
    ) -> float:
        m = norm.ppf((1 + significance) / 2) + norm.ppf(power)
        if not test_data:
            raise ValueError("test_data is required")

        n_test, n_control = len(test_data), len(data)

        var_test, var_control = test_data.var(ddof=1), data.var(ddof=1)
        s = np.sqrt(var_test / n_test + var_control / n_control)

        return m * s


#
#
# class StatPowerByTTestInd(TestPower):
#
#     def _inner_function(self, control_data, test_data) -> ExperimentData:
#         control_size = len(control_data)
#         test_size = len(test_data)
#
#         analysis = TTestIndPower()
#         ratio = test_size / control_size
#         return analysis.power(
#             effect_size=effect_size,
#             nobs1=test_size,
#             ratio=ratio,
#             alpha=significance,
#
from typing import Optional, Any, Dict, Union, Sequence

import numpy as np
from scipy.stats import norm

from ..comparators.abstract import Comparator
from ..dataset import ABCRole, Dataset, ExperimentData, TargetRole
from ..utils import SpaceEnum


# class MDEBySize(GroupComparator):
#     def __init__(
#         self,
#         grouping_role: Optional[ABCRole] = None,
#         space: SpaceEnum = SpaceEnum.auto,
#         full_name: Optional[str] = None,
#         key: Any = "",
#         power: float = 0.8,
#         significance: float = 0.95,
#     ):
#         super().__init__(grouping_role, space, full_name, key)
#         self.power = power
#         self.significance = significance
#
#     @staticmethod
#     def _inner_function(
#         control_data, test_data, significance=0.95, power=0.8, **kwargs
#     ) -> Dict[str, Any]:
#         result = {}
#         m = norm.ppf(1 - significance / 2) - norm.ppf(power)
#         n_control, n_test = len(control_data), len(test_data)
#         proportion = n_test / (n_test + n_control)
#         p = np.sqrt(1 / (proportion * (1 - proportion)))
#         for target in control_data.columns:
#             var_control = control_data[target].var()
#             var_test = test_data[target].var()
#             s = np.sqrt(var_test / n_test + var_control / n_control)
#             result[target] = p * m * s
#
#         return result
#
#     @staticmethod
#     def calc(
#         cls: Dataset,
#         data: Union[Sequence[str], str, None],
#         group_field: Optional[str] = None,
#         grouping_data=None,
#         target_fields=None,
#         **kwargs
#     ):
#         return GroupComparator.calc(
#             data=data,
#             group_field=group_field,
#             target_fields=target_fields,
#             comparison_function=MDEBySize._inner_function,
#             power=power,
#             significance=target_fields,
#         )
#
#     def execute(self, data: ExperimentData) -> ExperimentData:
#         subdata = data.ds.loc[
#             :, data.ds.get_columns_by_roles([TargetRole(), self.grouping_role])
#         ]
#         ed = super().execute(ExperimentData(subdata))
#         return self._set_value(data, ed.analysis_tables[self._id])



from typing import Optional, Any, Dict

from ..comparators.distances import MahalanobisDistance
from ..dataset import (
    Dataset,
    ABCRole,
    FeatureRole,
    ExperimentData,
    AdditionalMatchingRole,
)
from ..executor import MLExecutor
from ..extensions.faiss import FaissExtension
from ..utils import SpaceEnum, ExperimentDataEnum
from ..utils.errors import PairsNotFoundError


class FaissNearestNeighbors(MLExecutor):
    def __init__(
        self,
        n_neighbors: int = 1,
        two_sides: bool = False,
        test_pairs: bool = False,
        grouping_role: Optional[ABCRole] = None,
        key: Any = "",
    ):
        self.n_neighbors = n_neighbors
        self.two_sides = two_sides
        self.test_pairs = test_pairs
        super().__init__(
            grouping_role=grouping_role, target_role=FeatureRole(), key=key
        )

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_field: Optional[str] = None,
        n_neighbors: Optional[int] = None,
        two_sides: Optional[bool] = None,
        test_pairs: Optional[bool] = None,
        **kwargs,
    ) -> Dict:
        if test_pairs is not True:
            data = cls._inner_function(
                data=grouping_data[0][1],
                test_data=grouping_data[1][1],
                n_neighbors=n_neighbors or 1,
                **kwargs,
            )
            if two_sides is not True:
                return {"test": data}
            return {
                "test": data,
                "control": cls._inner_function(
                    data=grouping_data[1][1],
                    test_data=grouping_data[0][1],
                    n_neighbors=n_neighbors or 1,
                    **kwargs,
                ),
            }
        data = cls._inner_function(
            data=grouping_data[1][1],
            test_data=grouping_data[0][1],
            n_neighbors=n_neighbors or 1,
            **kwargs,
        )
        if two_sides is not True:
            return {"control": data}
        return {
            "control": data,
            "test": cls._inner_function(
                data=grouping_data[1][1],
                test_data=grouping_data[0][1],
                n_neighbors=n_neighbors or 1,
                **kwargs,
            ),
        }

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        target_data: Optional[Dataset] = None,
        n_neighbors: Optional[int] = None,
        **kwargs,
    ) -> Any:
        return FaissExtension(n_neighbors=n_neighbors or 1).calc(
            data=data, test_data=test_data
        )

    def fit(self, X: Dataset, Y: Optional[Dataset] = None) -> "MLExecutor":
        return FaissExtension(self.n_neighbors).fit(X=X, Y=Y)

    def predict(self, X: Dataset) -> Dataset:
        return FaissExtension().predict(X)

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, features_fields = self._get_fields(data=data)
        if group_field[0] in data.groups:
            grouping_data = list(data.groups[group_field[0]].items())
        else:
            grouping_data = data.ds.groupby(group_field, fields_list=features_fields)
        distances_keys = data.get_ids(MahalanobisDistance, ExperimentDataEnum.groups)
        if len(distances_keys["MahalanobisDistance"]["groups"]) > 0:
            grouping_data = list(
                data.groups[distances_keys["MahalanobisDistance"]["groups"][0]].items()
            )
        compare_result = self.calc(
            data=data.ds,
            group_field=group_field,
            grouping_data=grouping_data,
            features_fields=features_fields,
            n_neighbors=self.n_neighbors,
            two_sides=self.two_sides,
            test_pairs=self.test_pairs,
        )
        ds = data.ds.groupby(group_field)
        matched_indexes = Dataset.create_empty()
        for i in range(len(compare_result.columns)):
            group = (
                grouping_data[1][1]
                if compare_result.columns[i] == "test"
                else grouping_data[0][1]
            )
            t_ds = ds[0][1] if compare_result.columns[i] == "test" else ds[1][1]
            t_index_field = (
                compare_result[compare_result.columns[i]]
                .loc[: len(group) - 1]
                .rename({compare_result.columns[i]: "indexes"})
            )
            if t_index_field.isna().sum() > 0:
                raise PairsNotFoundError
            matched_indexes = matched_indexes.append(
                Dataset.from_dict(
                    data={
                        "indexes": t_ds.iloc[
                            list(map(lambda x: int(x[0]), t_index_field.get_values()))
                        ].index
                    },
                    roles={"indexes": AdditionalMatchingRole()},
                    index=group.index,
                )
            ).sort()
        if len(matched_indexes) < len(data.ds) and not self.two_sides:
            matched_indexes = matched_indexes.reindex(data.ds.index, fill_value=-1)
        elif len(matched_indexes) < len(data.ds) and self.two_sides:
            raise PairsNotFoundError
        return self._set_value(data, matched_indexes, key="matched")



from .shuffle import Shuffle
from .filters import CVFilter, ConstFilter, NanFilter, CorrFilter, OutliersFilter
from .category_agg import CategoryAggregator
from .na_filler import NaFiller
from ..encoders.encoders import DummyEncoder

__all__ = [
    "Shuffle",
    "CVFilter",
    "ConstFilter",
    "NanFilter",
    "CorrFilter",
    "OutliersFilter",
    "CategoryAggregator",
    "DummyEncoder",
    "NaFilter",
]



from typing import Dict, Any

from ..analyzers.ab import ABAnalyzer
from ..dataset import ExperimentData, Dataset
from ..utils import ExperimentDataEnum
from .aa import OneAADictReporter
from ..comparators import TTest, UTest, Chi2Test


class ABDictReporter(OneAADictReporter):
    def extract_analyzer_data(self, data: ExperimentData) -> Dict[str, Any]:
        analyzer_id = data.get_one_id(ABAnalyzer, ExperimentDataEnum.analysis_tables)
        return self.extract_from_one_row_dataset(data.analysis_tables[analyzer_id])

    def extract_tests(self, data: ExperimentData) -> Dict[str, Any]:
        test_ids = data.get_ids(
            [TTest, UTest, Chi2Test], searched_space=ExperimentDataEnum.analysis_tables
        )
        result = {}
        for class_, ids in test_ids.items():
            result.update(
                self._extract_from_comparators(
                    data, ids[ExperimentDataEnum.analysis_tables.value]
                )
            )
        return {k: v for k, v in result.items() if "pass" in k or "p-value" in k}

    def extract_data_from_analysis_tables(self, data: ExperimentData) -> Dict[str, Any]:
        result = {}
        result.update(self.extract_group_sizes(data))
        result.update(self.extract_group_difference(data))
        result.update(self.extract_tests(data))
        result.update(self.extract_analyzer_data(data))
        return result

    def report(self, data: ExperimentData) -> Dict[str, Any]:
        result = {}
        result.update(self.extract_data_from_analysis_tables(data))
        return result


class ABDatasetReporter(ABDictReporter):
    @staticmethod
    def _invert_aa_format(table: Dataset) -> Dataset:
        return table.replace("NOT OK", "N").replace("OK", "NOT OK").replace("N", "OK")

    def report(self, data: ExperimentData):
        front_buffer = self.front
        self.front = False
        dict_report = super().report(data)
        self.front = front_buffer
        result = self.convert_flat_dataset(dict_report)
        return self._invert_aa_format(result)



from typing import Dict, Any, Union

from ..dataset import ExperimentData, Dataset
from .aa import OneAADictReporter
from .abstract import DatasetReporter


class HomoDictReporter(OneAADictReporter):
    def report(self, data: ExperimentData) -> Dict[str, Any]:
        return self.extract_data_from_analysis_tables(data)


class HomoDatasetReporter(DatasetReporter):
    def __init__(self):
        super().__init__(dict_reporter=HomoDictReporter(front=False))

    @staticmethod
    def convert_to_dataset(data: Dict) -> Union[Dict[str, Dataset], Dataset]:
        return HomoDictReporter.convert_flat_dataset(data)



"""
This module provides the core Dataset and ExperimentData classes for data manipulation and analysis.

The Dataset class extends DatasetBase to provide a rich interface for working with tabular data,
including support for various data operations, statistical functions, and role-based column management.

The ExperimentData class provides a container for organizing experimental data, analysis results,
and metadata.

Classes:
    Dataset: Main class for working with tabular data with role-based column management
    ExperimentData: Container class for organizing experimental data and results
    DatasetAdapter: Utility class for converting various data types to Dataset objects

Example:
    >>> roles = {'col1': InfoRole(), 'col2': InfoRole()}
    >>> data = pd.DataFrame({'col1': [1,2,3], 'col2': [4,5,6]})
    >>> ds = Dataset(roles=roles, data=data)
    >>> ds.mean()
    col1    2.0
    col2    5.0
"""

import warnings
from copy import copy, deepcopy
from collections.abc import Iterable
from typing import (
    Union,
    List,
    Iterable,
    Any,
    Dict,
    Callable,
    Hashable,
    Optional,
    Tuple,
    Literal,
    Sequence,
)

import pandas as pd  # type: ignore

from ..utils import (
    ID_SPLIT_SYMBOL,
    BackendsEnum,
    ConcatBackendError,
    ConcatDataError,
    ExperimentDataEnum,
    FromDictTypes,
    MultiFieldKeyTypes,
    NotFoundInExperimentDataError,
    DataTypeError,
    BackendTypeError,
    ScalarType,
)
from .abstract import DatasetBase
from .roles import (
    StatisticRole,
    InfoRole,
    ABCRole,
    FilterRole,
    DefaultRole,
    AdditionalRole,
)
from ..utils.adapter import Adapter
from ..utils.errors import InvalidArgumentError


class Dataset(DatasetBase):
    """
    A class for working with tabular data that extends DatasetBase with additional functionality.

    The Dataset class provides a rich interface for data manipulation, statistical operations,
    and role-based column management. It supports multiple backends (currently pandas)
    and provides consistent access patterns regardless of the underlying data storage.

    Attributes:
        loc (Locker): Label-based indexer for accessing data
        iloc (ILocker): Integer-based indexer for accessing data
        roles (Dict[str, ABCRole]): Mapping of column names to their roles
        data (pd.DataFrame): The underlying data storage

    Args:
        roles (Union[Dict[ABCRole, Union[List[str], str]], Dict[str, ABCRole]]):
            Role definitions for columns
        data (Optional[Union[pd.DataFrame, str]]): Input data or path to data
        backend (Optional[BackendsEnum]): Backend storage type
        default_role (Optional[ABCRole]): Default role for columns without explicit roles
    """

    class Locker:
        """
        Label-based indexer for accessing data in the Dataset.

        Provides a pandas-like .loc[] interface for label-based indexing.
        """

        def __init__(self, backend, roles):
            self.backend = backend
            self.roles = roles

        def __getitem__(self, item) -> "Dataset":
            """
            Get data subset using label-based indexing.
            """
            t_data = self.backend.loc(item)
            return Dataset(
                data=t_data,
                roles={k: v for k, v in self.roles.items() if k in t_data.columns},
            )

        def __setitem__(self, item, value):
            """
            Set data values using label-based indexing.
            """
            column_name = item[1]
            column_data_type = self.roles[column_name].data_type
            # Check if value matches expected data type
            if (
                column_data_type == None
                or (
                    isinstance(value, Iterable)
                    and all(isinstance(v, column_data_type) for v in value)
                )
                or isinstance(value, column_data_type)
                or value is None
            ):
                if column_name not in self.backend.data.columns:
                    raise KeyError("Column must be added by using add_column method.")
                else:
                    self.backend.data.loc[item] = value
            else:
                raise TypeError("Value type does not match the expected data type.")

    class ILocker:
        """
        Integer-based indexer for accessing data in the Dataset.

        Provides a pandas-like .iloc[] interface for integer-based indexing.
        """

        def __init__(self, backend, roles):
            self.backend = backend
            self.roles = roles

        def __getitem__(self, item) -> "Dataset":
            """
            Get data subset using integer-based indexing.
            """
            t_data = self.backend.iloc(item)
            return Dataset(
                data=t_data,
                roles={k: v for k, v in self.roles.items() if k in t_data.columns},
            )

        def __setitem__(self, item, value):
            """
            Set data values using integer-based indexing.
            """
            column_index = item[1]
            column_name = self.backend.data.columns[column_index]
            column_data_type = self.roles[column_name].data_type
            # Check if value matches expected data type
            if (
                column_data_type == None
                or (
                    isinstance(value, Iterable)
                    and all(isinstance(v, column_data_type) for v in value)
                )  # check for backend specific list (?)
                or isinstance(value, column_data_type)
            ):
                if column_index >= len(self.backend.data.columns):
                    raise IndexError("Column must be added by using add_column method.")
                else:
                    self.backend.data.iloc[item] = value
            else:
                raise TypeError("Value type does not match the expected data type.")

    def __init__(
        self,
        roles: Union[
            Dict[ABCRole, Union[List[str], str]],
            Dict[str, ABCRole],
        ],
        data: Optional[Union[pd.DataFrame, str]] = None,
        backend: Optional[BackendsEnum] = None,
        default_role: Optional[ABCRole] = None,
    ):
        """
        Initialize a new Dataset instance.

        Args:
            roles: Role definitions for columns
            data: Input data or path to data
            backend: Backend storage type
            default_role: Default role for columns without explicit roles
        """
        super().__init__(roles, data, backend, default_role)
        self.loc = self.Locker(self._backend, self.roles)
        self.iloc = self.ILocker(self._backend, self.roles)

    def __getitem__(self, item: Union[Iterable, str, int]) -> "Dataset":
        """
        Get a subset of the dataset by column selection.

        Args:
            item: Column name(s) to select

        Returns:
            Dataset containing only the selected columns
        """
        if isinstance(item, Dataset):
            item = item.data
        items = (
            [item] if isinstance(item, str) or not isinstance(item, Iterable) else item
        )
        roles: Dict = {
            column: (
                self.roles[column]
                if column in self.columns and self.roles.get(column, False)
                else InfoRole()
            )
            for column in items
        }
        result = Dataset(data=self._backend.__getitem__(item), roles=roles)
        result.tmp_roles = {
            key: value for key, value in self.tmp_roles.items() if key in items
        }
        return result

    def __setitem__(self, key: str, value: Any):
        """
        Set values for a column in the dataset.

        Args:
            key: Column name
            value: Values to set

        Raises:
            TypeError: If value type doesn't match column's expected type
        """
        if isinstance(value, Dataset):
            value = value.data
        if key not in self.columns and isinstance(key, str):
            self.add_column(value, {key: InfoRole()})
            warnings.warn(
                "Column must be added by using add_column method.",
                category=SyntaxWarning,
            )
            self.data[key] = value
        else:
            column_data_type = self.roles[key].data_type
            if (
                column_data_type == None
                or (
                    isinstance(value, Iterable)
                    and all(isinstance(v, column_data_type) for v in value)
                )  # check for backend specific list (?)
                or isinstance(value, column_data_type)
            ):
                self.data[key] = value
            else:
                raise TypeError("Value type does not match the expected data type.")

    def __binary_magic_operator(self, other, func_name: str) -> Any:
        """
        Helper method for implementing binary operators.

        Args:
            other: Right-hand operand
            func_name: Name of operator function to call

        Returns:
            Result of binary operation

        Raises:
            DataTypeError: If other has invalid type
            BackendTypeError: If backends don't match
        """
        if not any(
            isinstance(other, t) for t in [Dataset, str, int, float, bool, Sequence]
        ):
            raise DataTypeError(type(other))
        func = getattr(self._backend, func_name)
        t_roles = deepcopy(self.roles)
        for role in t_roles.values():
            role.data_type = None
        if isinstance(other, Dataset):
            if type(other._backend) is not type(self._backend):
                raise BackendTypeError(type(other._backend), type(self._backend))
            other = other.rename(
                {
                    other.columns[i]: self.data.columns[i]
                    for i in range(len(other.columns))
                }
            ).backend
        return Dataset(roles=t_roles, data=func(other))

    # comparison operators:
    def __eq__(self, other):
        """
        Implement equality comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__eq__")

    def __ne__(self, other):
        """
        Implement inequality comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__ne__")

    def __le__(self, other):
        """
        Implement less than or equal comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__le__")

    def __lt__(self, other):
        """
        Implement less than comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__lt__")

    def __ge__(self, other):
        """
        Implement greater than or equal comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__ge__")

    def __gt__(self, other):
        """
        Implement greater than comparison.
        """
        return self.__binary_magic_operator(other=other, func_name="__gt__")

    # unary operators:
    def __pos__(self):
        """
        Implement unary positive.
        """
        return Dataset(roles=self.roles, data=(+self._backend))

    def __neg__(self):
        """
        Implement unary negation.
        """
        return Dataset(roles=self.roles, data=(-self._backend))

    def __abs__(self):
        """
        Implement absolute value.
        """
        return Dataset(roles=self.roles, data=abs(self._backend))

    def __invert__(self):
        """
        Implement bitwise inversion.
        """
        return Dataset(roles=self.roles, data=(~self._backend))

    def __round__(self, ndigits: int = 0):
        """
        Implement rounding.
        """
        return Dataset(roles=self.roles, data=round(self._backend, ndigits))

    def __bool__(self):
        """
        Implement truth value testing.
        """
        return not self._backend.is_empty()

    # Binary math operators:
    def __add__(self, other):
        """
        Implement addition.
        """
        return self.__binary_magic_operator(other=other, func_name="__add__")

    def __sub__(self, other):
        """
        Implement subtraction.
        """
        return self.__binary_magic_operator(other=other, func_name="__sub__")

    def __mul__(self, other):
        """
        Implement multiplication.
        """
        return self.__binary_magic_operator(other=other, func_name="__mul__")

    def __floordiv__(self, other):
        """
        Implement floor division.
        """
        return self.__binary_magic_operator(other=other, func_name="__floordiv__")

    def __div__(self, other):
        """
        Implement division.
        """
        return self.__binary_magic_operator(other=other, func_name="__div__")

    def __truediv__(self, other):
        """
        Implement true division.
        """
        return self.__binary_magic_operator(other=other, func_name="__truediv__")

    def __mod__(self, other):
        """
        Implement modulo.
        """
        return self.__binary_magic_operator(other=other, func_name="__mod__")

    def __pow__(self, other):
        """
        Implement exponentiation.
        """
        return self.__binary_magic_operator(other=other, func_name="__pow__")

    def __and__(self, other):
        """
        Implement bitwise and.
        """
        return self.__binary_magic_operator(other=other, func_name="__and__")

    def __or__(self, other):
        """
        Implement bitwise or.
        """
        return self.__binary_magic_operator(other=other, func_name="__or__")

    # Right math operators:
    def __radd__(self, other):
        """
        Implement reverse addition.
        """
        return self.__binary_magic_operator(other=other, func_name="__radd__")

    def __rsub__(self, other):
        """
        Implement reverse subtraction.
        """
        return self.__binary_magic_operator(other=other, func_name="__rsub__")

    def __rmul__(self, other):
        """
        Implement reverse multiplication.
        """
        return self.__binary_magic_operator(other=other, func_name="__rmul__")

    def __rfloordiv__(self, other):
        """
        Implement reverse floor division.
        """
        return self.__binary_magic_operator(other=other, func_name="__rfloordiv__")

    def __rdiv__(self, other):
        """
        Implement reverse division.
        """
        return self.__binary_magic_operator(other=other, func_name="__rdiv__")

    def __rtruediv__(self, other):
        """
        Implement reverse true division.
        """
        return self.__binary_magic_operator(other=other, func_name="__rtruediv__")

    def __rmod__(self, other):
        """
        Implement reverse modulo.
        """
        return self.__binary_magic_operator(other=other, func_name="__rmod__")

    def __rpow__(self, other) -> Any:
        """
        Implement reverse exponentiation.
        """
        return self.__binary_magic_operator(other=other, func_name="__rpow__")

    @property
    def index(self):
        """
        Get the index of the dataset.
        """
        return self.backend.index

    @index.setter
    def index(self, value):
        """
        Set the index of the dataset.
        """
        self.backend.data.index = value

    @property
    def data(self):
        """
        Get the underlying data.
        """
        return self._backend.data

    @data.setter
    def data(self, value):
        """
        Set the underlying data.
        """
        self.backend.data = value

    @property
    def columns(self):
        """
        Get the column names.
        """
        return self.backend.columns

    @staticmethod
    def create_empty(roles=None, index=None, backend=BackendsEnum.pandas) -> "Dataset":
        """
        Create an empty dataset.

        Args:
            roles: Role definitions for columns
            index: Index for the empty dataset
            backend: Backend storage type

        Returns:
            Empty Dataset instance
        """
        if roles is None:
            roles = {}
        index = [] if index is None else index
        columns = list(roles.keys())
        ds = Dataset(roles=roles, backend=backend)
        ds._backend = ds._backend.create_empty(index, columns)
        ds.data = ds.backend.data
        return ds

    def _convert_data_after_agg(self, result) -> Union["Dataset", float]:
        """
        Convert aggregation result to appropriate type.

        Args:
            result: Result of aggregation operation

        Returns:
            Dataset or float depending on result type
        """
        if isinstance(result, float):
            return result
        role: ABCRole = StatisticRole()
        return Dataset(data=result, roles={column: role for column in result.columns})

    def add_column(
        self,
        data,
        role: Optional[Dict[str, ABCRole]] = None,
        index: Optional[Iterable[Hashable]] = None,
    ):
        """
        Add a new column to the dataset.

        Args:
            data: Column data to add
            role: Role definition for the new column
            index: Index for the new column

        Returns:
            Self for method chaining

        Raises:
            ValueError: If role is None and data is not a Dataset
        """
        if role is None:
            if not isinstance(data, Dataset):
                raise ValueError("If role is None, data must be a Dataset")
            if any([col in self.columns for col in data.columns]):
                raise ValueError("Columns with the same name already exist")
            self.roles.update(data.roles)
            self._backend.add_column(
                data.data,
                data.columns,
                index,
            )
        else:
            if any([col in self.columns for col in role.keys()]):
                raise ValueError("Columns with the same name already exist")
            if isinstance(role, Dict) and any(
                [not isinstance(r, ABCRole) for r in role.values()]
            ):
                raise TypeError("Role values must be of type ABCRole")
            if isinstance(data, Dataset):
                data = data.data
            self.roles.update(role)
            self._backend.add_column(data, list(role.keys()), index)
        return self

    def _check_other_dataset(self, other):
        """
        Check compatibility with another dataset.

        Args:
            other: Dataset to check compatibility with

        Raises:
            ConcatDataError: If other is not a Dataset
            ConcatBackendError: If backends don't match
        """
        if not isinstance(other, Dataset):
            raise ConcatDataError(type(other))
        if type(other._backend) is not type(self._backend):
            raise ConcatBackendError(type(other._backend), type(self._backend))

    def astype(
        self, dtype: Dict[str, type], errors: Literal["raise", "ignore"] = "raise"
    ) -> "Dataset":
        """
        Change the data type of one or more columns.

        Parameters:
        - dtype: Dictionary where keys are column names and values are target types.
        - errors: If 'raise', raises an error on invalid types; if 'ignore', skips invalid types.

        Returns:
        - A new Dataset object with the specified data types applied.
        """

        for col, _ in dtype.items():
            if (errors == "raise") and (col not in self.columns):
                raise KeyError(f"Column '{col}' does not exist in the Dataset.")

        new_backend = deepcopy(self._backend)
        new_backend.data = new_backend.astype(dtype, errors)
        new_roles = deepcopy(self.roles)

        if errors == "ignore":
            for col, target_type in dtype.items():
                if new_backend.get_column_type(col) == target_type:
                    new_roles[col].data_type = target_type
        elif errors == "raise":
            for col, target_type in dtype.items():
                new_roles[col].data_type = target_type

        return Dataset(roles=new_roles, data=new_backend.data)

    def append(self, other, reset_index=False, axis=0) -> "Dataset":
        """
        Append rows or columns from another dataset.

        Args:
            other: Dataset to append
            reset_index: Whether to reset index after append
            axis: 0 for row append, 1 for column append

        Returns:
            New Dataset with appended data
        """
        other = Adapter.to_list(other)

        new_roles = deepcopy(self.roles)
        for o in other:
            self._check_other_dataset(o)
            new_roles.update(o.roles)

        return Dataset(
            roles=new_roles, data=self.backend.append(other, reset_index, axis)
        )

    # TODO: set backend by backend object
    @staticmethod
    def from_dict(
        data: FromDictTypes,
        roles: Union[
            Dict[ABCRole, Union[List[str], str]],
            Dict[str, ABCRole],
        ],
        backend: BackendsEnum = BackendsEnum.pandas,
        index=None,
    ) -> "Dataset":
        """
        Create a Dataset from a dictionary.

        Args:
            data: Dictionary containing data
            roles: Role definitions for columns
            backend: Backend storage type
            index: Index for the dataset

        Returns:
            New Dataset created from dictionary
        """
        ds = Dataset(roles=roles, backend=backend)
        ds._backend = ds._backend.from_dict(data, index)
        ds.data = ds._backend.data
        return ds

    # What is going to happen when a matrix is returned?
    def apply(
        self,
        func: Callable,
        role: Dict[str, ABCRole],
        axis: int = 0,
        **kwargs,
    ) -> "Dataset":
        """
        Apply a function to the dataset.

        Args:
            func: Function to apply
            role: Role definition for result
            axis: Axis to apply function along
            **kwargs: Additional arguments for func

        Returns:
            New Dataset with function applied
        """
        if self.is_empty():
            return deepcopy(self)
        tmp_data = self._backend.apply(
            func=func, axis=axis, column_name=list(role.keys())[0], **kwargs
        )
        tmp_roles = (
            {list(role.keys())[0]: list(role.values())[0]}
            if ((not tmp_data.any().any()) and len(role) > 1)
            else role
        )
        return Dataset(
            data=tmp_data,
            roles=tmp_roles,
        )

    def map(self, func, na_action=None, **kwargs) -> "Dataset":
        """
        Apply a mapping function to the dataset.

        Args:
            func: Mapping function to apply
            na_action: How to handle NA values
            **kwargs: Additional arguments for func

        Returns:
            New Dataset with mapping applied
        """
        return Dataset(
            roles=self.roles,
            data=self._backend.map(func=func, na_action=na_action, **kwargs),
        )

    def is_empty(self) -> bool:
        """
        Check if dataset is empty.
        """
        return self._backend.is_empty()

    def unique(self) -> Dict[str, List[Any]]:
        """
        Get unique values for each column.
        """
        return self._backend.unique()

    def nunique(self, dropna: bool = False) -> Dict[str, int]:
        """
        Count unique values for each column.

        Args:
            dropna: Whether to exclude NA values

        Returns:
            Dictionary mapping column names to counts
        """
        return self._backend.nunique(dropna)

    def isin(self, values: Iterable) -> "Dataset":
        """
        Check whether values are contained in the dataset.

        Args:
            values: Values to check for

        Returns:
            Boolean mask as Dataset
        """
        role: ABCRole = FilterRole()
        return Dataset(
            roles={column: role for column in self.roles.keys()},
            data=self._backend.isin(values),
        )

    def groupby(
        self,
        by: Any,
        func: Optional[Union[str, List]] = None,
        fields_list: Optional[Union[str, List]] = None,
        **kwargs,
    ) -> List[Tuple[str, "Dataset"]]:
        """
        Group dataset by values.

        Args:
            by: Column(s) to group by
            func: Aggregation function to apply
            fields_list: Columns to include in result
            **kwargs: Additional arguments for groupby

        Returns:
            List of (group_key, group_data) tuples
        """
        if isinstance(by, Dataset) and len(by.columns) == 1:
            self.data = self.data.reset_index(drop=True)
            datasets = [
                (group, Dataset(roles=self.roles, data=self.data.loc[group_data.index]))
                for group, group_data in by._backend.groupby(by=by.columns[0], **kwargs)
            ]
        else:
            datasets = [
                (group, Dataset(roles=self.roles, data=data))
                for group, data in self._backend.groupby(by=by, **kwargs)
            ]
        if fields_list:
            fields_list = Adapter.to_list(fields_list)
            datasets = [(i, data[fields_list]) for i, data in datasets]
        if func:
            datasets = [(i, data.agg(func)) for i, data in datasets]
        for dataset in datasets:
            if isinstance(dataset, Dataset):
                dataset[1].tmp_roles = self.tmp_roles
        return datasets

    def sort(
        self,
        by: Optional[MultiFieldKeyTypes] = None,
        ascending: bool = True,
        **kwargs,
    ):
        """
        Sort dataset by values.

        Args:
            by: Column(s) to sort by
            ascending: Sort order
            **kwargs: Additional arguments for sort

        Returns:
            Sorted Dataset
        """
        if by is None:
            return Dataset(
                roles=self.roles,
                data=self.backend.sort_index(ascending=ascending, **kwargs),
            )
        return Dataset(
            roles=self.roles,
            data=self.backend.sort_values(by=by, ascending=ascending, **kwargs),
        )

    def fillna(
        self,
        values: Union[ScalarType, Dict[str, ScalarType], None] = None,
        method: Optional[Literal["bfill", "ffill"]] = None,
        **kwargs,
    ):
        """
        Fill NA values.

        Args:
            values: Values to fill with
            method: Method for filling
            **kwargs: Additional arguments

        Returns:
            Dataset with NA values filled

        Raises:
            ValueError: If neither values nor method provided
        """
        if values is None and method is None:
            raise ValueError("Value or filling method must be provided")
        return Dataset(
            roles=self.roles,
            data=self.backend.fillna(values=values, method=method, **kwargs),
        )

    def mean(self):
        """
        Calculate mean of numeric columns.
        """
        return self._convert_data_after_agg(self._backend.mean())

    def max(self):
        """
        Calculate maximum of columns.
        """
        return self._convert_data_after_agg(self._backend.max())

    def reindex(self, labels, fill_value: Optional[Any] = None) -> "Dataset":
        """
        Reindex the dataset.

        Args:
            labels: New index labels
            fill_value: Value to use for missing values

        Returns:
            Reindexed Dataset
        """
        return Dataset(
            self.roles, data=self.backend.reindex(labels, fill_value=fill_value)
        )

    def idxmax(self):
        """
        Get index of maximum values.
        """
        return self._convert_data_after_agg(self._backend.idxmax())

    def min(self):
        """
        Calculate minimum of columns.
        """
        return self._convert_data_after_agg(self._backend.min())

    def count(self):
        """
        Count non-NA values.
        """
        if self.is_empty():
            return Dataset.create_empty({role: InfoRole() for role in self.roles})
        return self._convert_data_after_agg(self._backend.count())

    def sum(self):
        """
        Calculate sum of columns.
        """
        return self._convert_data_after_agg(self._backend.sum())

    def log(self):
        """
        Calculate natural logarithm.
        """
        return self._convert_data_after_agg(self._backend.log())

    def mode(self, numeric_only: bool = False, dropna: bool = True):
        """
        Calculate mode of columns.

        Args:
            numeric_only: Whether to include only numeric columns
            dropna: Whether to exclude NA values

        Returns:
            Mode values as Dataset
        """
        t_data = self._backend.mode(numeric_only=numeric_only, dropna=dropna)
        return Dataset(data=t_data, roles={role: InfoRole() for role in t_data.columns})

    def var(self, skipna: bool = True, ddof: int = 1, numeric_only: bool = False):
        """
        Calculate variance of columns.

        Args:
            skipna: Whether to exclude NA values
            ddof: Delta degrees of freedom
            numeric_only: Whether to include only numeric columns

        Returns:
            Variance values as Dataset
        """
        return self._convert_data_after_agg(
            self._backend.var(skipna=skipna, ddof=ddof, numeric_only=numeric_only)
        )

    def agg(self, func: Union[str, List]):
        """
        Aggregate using one or more operations.

        Args:
            func: Function or list of functions to apply

        Returns:
            Aggregated results as Dataset
        """
        return self._convert_data_after_agg(self._backend.agg(func))

    def std(self, skipna: bool = True, ddof: int = 1):
        """
        Calculate standard deviation.
        """
        return self._convert_data_after_agg(self._backend.std(skipna=skipna, ddof=ddof))

    def quantile(self, q: float = 0.5):
        """
        Calculate quantiles for each column.

        Args:
            q: Quantile to compute, between 0 and 1

        Returns:
            Quantile values for each column
        """
        return self._convert_data_after_agg(self._backend.quantile(q=q))

    def coefficient_of_variation(self):
        """
        Calculate coefficient of variation.
        """
        return self._convert_data_after_agg(self._backend.coefficient_of_variation())

    def corr(self, method="pearson", numeric_only=False):
        """
        Calculate correlation between columns.

        Args:
            method: Correlation method
            numeric_only: Whether to include only numeric columns

        Returns:
            Correlation matrix as Dataset
        """
        t_data = self._backend.corr(method=method, numeric_only=numeric_only)
        t_roles = {column: self.roles[column] for column in t_data.columns}
        return Dataset(roles=t_roles, data=t_data)

    def value_counts(
        self,
        normalize: bool = False,
        sort: bool = True,
        ascending: bool = False,
        dropna: bool = True,
    ):
        """
        Count unique values.

        Args:
            normalize: Return proportions instead of counts
            sort: Sort by counts
            ascending: Sort order
            dropna: Whether to exclude NA values

        Returns:
            Value counts as Dataset
        """
        t_data = self._backend.value_counts(
            normalize=normalize, sort=sort, ascending=ascending, dropna=dropna
        )
        t_roles = deepcopy(self.roles)
        column_name = "proportion" if normalize else "count"
        if column_name not in t_data:
            t_data = t_data.rename(columns={0: column_name})
        t_roles[column_name] = StatisticRole()
        return Dataset(roles=t_roles, data=t_data)

    def na_counts(self):
        """Count NA values"""
        return self._convert_data_after_agg(self._backend.na_counts())

    def dropna(
        self,
        how: Literal["any", "all"] = "any",
        subset: Union[str, Iterable[str], None] = None,
        axis: Union[Literal["index", "rows", "columns"], int] = 0,
    ):
        """
        Remove missing values from the dataset.

        Args:
            how: How to drop rows/columns with missing values.
                'any': Drop if any NA values are present
                'all': Drop only if all values are NA
            subset: Labels of columns to check for missing values, or None to use all columns
            axis: Which axis to drop values from
                0/'index'/'rows': Drop rows
                1/'columns': Drop columns

        Returns:
            Dataset: A new dataset with NA values removed according to specified criteria
        """
        # Drop NA values using backend implementation
        new_data = self._backend.dropna(how=how, subset=subset, axis=axis)

        # Update roles based on axis - keep all roles for row drops, filter for column drops
        new_roles = (
            self.roles
            if axis == 0
            else {column: self.roles[column] for column in new_data.columns}
        )

        # Return new dataset with updated data and roles
        return Dataset(
            roles=new_roles,
            data=new_data,
        )

    def isna(self):
        """
        Check for missing values in the dataset.

        Returns:
            Dataset: A boolean dataset indicating missing values
        """
        return self._convert_data_after_agg(self._backend.isna())

    def select_dtypes(self, include: Any = None, exclude: Any = None):
        """
        Select columns based on their dtype.

        Args:
            include: Dtypes to include
            exclude: Dtypes to exclude

        Returns:
            Dataset: Dataset with selected dtypes
        """
        # Filter data by dtypes
        t_data = self._backend.select_dtypes(include=include, exclude=exclude)

        # Keep only roles for remaining columns
        t_roles = {k: v for k, v in self.roles.items() if k in t_data.columns}
        return Dataset(roles=t_roles, data=t_data)

    def merge(
        self,
        right,
        on: Optional[str] = None,
        left_on: Optional[str] = None,
        right_on: Optional[str] = None,
        left_index: bool = False,
        right_index: bool = False,
        suffixes: Tuple[str, str] = ("_x", "_y"),
        how: Literal["left", "right", "outer", "inner", "cross"] = "inner",
    ):
        """
        Merge this dataset with another dataset.

        Args:
            right: Right dataset to merge with
            on: Column name(s) to join on
            left_on: Column(s) from left dataset to join on
            right_on: Column(s) from right dataset to join on
            left_index: Use left index as join key
            right_index: Use right index as join key
            suffixes: Suffixes to add to overlapping column names
            how: Type of merge to perform

        Returns:
            Dataset: Merged dataset

        Raises:
            DataTypeError: If right is not a Dataset
            BackendTypeError: If backends don't match
        """
        # Default to index merge if no columns specified
        if not any([on, left_on, right_on, left_index, right_index]):
            left_index = True
            right_index = True

        # Validate input types
        if not isinstance(right, Dataset):
            raise DataTypeError(type(right))
        if type(right._backend) is not type(self._backend):
            raise BackendTypeError(type(right._backend), type(self._backend))

        # Perform merge operation
        t_data = self._backend.merge(
            right=right._backend,
            on=on,
            left_on=left_on,
            right_on=right_on,
            left_index=left_index,
            right_index=right_index,
            suffixes=suffixes,
            how=how,
        )

        # Combine roles from both datasets
        t_roles = deepcopy(self.roles)
        t_roles.update(right.roles)

        # Handle suffixed column roles
        for c in t_data.columns:
            if f"{c}".endswith(suffixes[0]) and c[: -len(suffixes[0])] in self.columns:
                t_roles[c] = self.roles[c[: -len(suffixes[0])]]
            if f"{c}".endswith(suffixes[1]) and c[: -len(suffixes[1])] in right.columns:
                t_roles[c] = right.roles[c[: -len(suffixes[1])]]

        # Create final roles dict with only merged columns
        new_roles = {c: t_roles[c] for c in t_data.columns}
        return Dataset(roles=new_roles, data=t_data)

    def drop(self, labels: Any = None, axis: int = 1):
        """
        Drop specified labels from rows or columns.

        Args:
            labels: Labels to drop
            axis: 0 for rows, 1 for columns

        Returns:
            Dataset: Dataset with specified labels dropped
        """
        # Convert Dataset labels to list of indices
        if isinstance(labels, Dataset):
            labels = list(labels.index)

        # Drop specified labels
        t_data = self._backend.drop(labels=labels, axis=axis)

        # Update roles based on axis
        t_roles = (
            deepcopy(self.roles)
            if axis == 0
            else {c: self.roles[c] for c in t_data.columns}
        )
        return Dataset(roles=t_roles, data=t_data)

    def filter(
        self,
        items: Optional[List] = None,
        like: Optional[str] = None,
        regex: Optional[str] = None,
        axis: Optional[int] = None,
    ) -> "Dataset":
        """
        Filter rows or columns using specified criteria.

        Args:
            items: List of items to include
            like: Include labels matching this substring
            regex: Include labels matching this regular expression
            axis: 0 for rows, 1 for columns

        Returns:
            Dataset: Filtered dataset
        """
        t_data = self._backend.filter(items=items, like=like, regex=regex, axis=axis)
        t_roles = {c: self.roles[c] for c in t_data.columns if c in self.roles.keys()}
        return Dataset(roles=t_roles, data=t_data)

    def dot(self, other: "Dataset") -> "Dataset":
        """
        Compute matrix multiplication with another dataset.

        Args:
            other: Dataset to multiply with

        Returns:
            Dataset: Result of matrix multiplication
        """
        return Dataset(roles=other.roles, data=self.backend.dot(other.backend))

    def transpose(
        self,
        roles: Optional[Union[Dict[str, ABCRole], List[str]]] = None,
    ) -> "Dataset":
        """
        Transpose the dataset.

        Args:
            roles: New roles for transposed columns

        Returns:
            Dataset: Transposed dataset
        """
        # Get role names if provided
        roles_names: List[Union[str, None]] = (
            list(roles.keys()) or [] if isinstance(roles, Dict) else roles
        )

        # Transpose data
        result_data = self.backend.transpose(roles_names)

        # Create default roles if none provided
        if roles is None or isinstance(roles, List):
            names = result_data.columns if roles is None else roles
            roles = {column: DefaultRole() for column in names}

        return Dataset(roles=roles, data=result_data)

    def sample(
        self,
        frac: Optional[float] = None,
        n: Optional[int] = None,
        random_state: Optional[int] = None,
    ) -> "Dataset":
        """
        Return a random sample of the dataset.

        Args:
            frac: Fraction of rows to sample
            n: Number of rows to sample
            random_state: Random seed for reproducibility

        Returns:
            Dataset: Random sample of rows
        """
        return Dataset(
            self.roles,
            data=self.backend.sample(frac=frac, n=n, random_state=random_state),
        )

    def cov(self):
        """
        Calculate covariance matrix.

        Returns:
            Dataset: Covariance matrix
        """
        t_data = self.backend.cov()
        return Dataset(
            {column: DefaultRole() for column in t_data.columns}, data=t_data
        )

    def rename(self, names: Dict[str, str]):
        """
        Rename columns.

        Args:
            names: Dictionary mapping old names to new names

        Returns:
            Dataset: Dataset with renamed columns
        """
        roles = {names.get(column, column): role for column, role in self.roles.items()}
        return Dataset(roles, data=self.backend.rename(names))

    def replace(
        self,
        to_replace: Any = None,
        value: Any = None,
        regex: bool = False,
    ) -> "Dataset":
        """
        Replace values in the dataset.

        Args:
            to_replace: Values to replace
            value: Value to replace with
            regex: Whether to interpret to_replace as regular expressions

        Returns:
            Dataset: Dataset with replaced values
        """
        return Dataset(
            self.roles,
            data=self._backend.replace(to_replace=to_replace, value=value, regex=regex),
        )


class ExperimentData:
    """
    Class for managing experimental data and analysis results.
    """

    def __init__(self, data: Dataset):
        """
        Initialize ExperimentData.

        Args:
            data: Base dataset for the experiment
        """
        self._data = data
        self.additional_fields = Dataset.create_empty(index=data.index)
        self.variables: Dict[str, Dict[str, Union[int, float]]] = {}
        self.groups: Dict[str, Dict[str, Dataset]] = {}
        self.analysis_tables: Dict[str, Dataset] = {}
        self.id_name_mapping: Dict[str, str] = {}

    @property
    def ds(self):
        """
        Get the base dataset.
        """
        return self._data

    @staticmethod
    def create_empty(
        roles=None, backend=BackendsEnum.pandas, index=None
    ) -> "ExperimentData":
        """
        Create empty ExperimentData instance.

        Args:
            roles: Roles for columns
            backend: Backend to use
            index: Index for empty dataset

        Returns:
            ExperimentData: Empty experiment data instance
        """
        ds = Dataset.create_empty(backend, roles, index)
        return ExperimentData(ds)

    def check_hash(self, executor_id: int, space: ExperimentDataEnum) -> bool:
        """
        Check if executor ID exists in specified space.

        Args:
            executor_id: ID to check
            space: Space to check in

        Returns:
            bool: Whether ID exists in space
        """
        if space == ExperimentDataEnum.additional_fields:
            return executor_id in self.additional_fields.columns
        elif space == ExperimentDataEnum.variables:
            return executor_id in self.variables.keys()
        elif space == ExperimentDataEnum.analysis_tables:
            return executor_id in self.analysis_tables
        else:
            return any(self.check_hash(executor_id, s) for s in ExperimentDataEnum)

    def set_value(
        self,
        space: ExperimentDataEnum,
        executor_id: Union[str, Dict[str, str]],
        value: Any,
        key: Optional[str] = None,
        role=None,
    ) -> "ExperimentData":
        """
        Set a value in the specified space.

        Args:
            space: Space to set value in
            executor_id: ID for the value
            value: Value to set
            key: Optional key for the value
            role: Optional role for the value

        Returns:
            ExperimentData: Self for chaining
        """
        # Handle additional fields
        if space == ExperimentDataEnum.additional_fields:
            if not isinstance(value, Dataset) or len(value.columns) == 1:
                self.additional_fields = self.additional_fields.add_column(
                    data=value, role={executor_id: role}
                )
            else:
                rename_dict = (
                    {value.columns[0]: executor_id}
                    if isinstance(executor_id, str)
                    else executor_id
                )
                value = value.rename(names=rename_dict)
                self.additional_fields = self.additional_fields.merge(
                    right=value, left_index=True, right_index=True
                )

        # Handle analysis tables
        elif space == ExperimentDataEnum.analysis_tables:
            self.analysis_tables[executor_id] = value

        # Handle variables
        elif space == ExperimentDataEnum.variables:
            if executor_id in self.variables:
                self.variables[executor_id][key] = value
            elif isinstance(value, Dict):
                self.variables[executor_id] = value
            else:
                self.variables[executor_id] = {key: value}

        # Handle groups
        elif space == ExperimentDataEnum.groups:
            if executor_id not in self.groups:
                self.groups[executor_id] = {key: value}
            else:
                self.groups[executor_id][key] = value

        return self

    def get_ids(
        self,
        classes: Union[type, Iterable[type], str, Iterable[str]],
        searched_space: Union[
            ExperimentDataEnum, Iterable[ExperimentDataEnum], None
        ] = None,
        key: Optional[str] = None,
    ) -> Dict[str, Dict[str, List[str]]]:
        """
        Get IDs matching specified criteria.

        Args:
            classes: Classes to search for
            searched_space: Spaces to search in
            key: Optional key to match

        Returns:
            Dict mapping classes to spaces and matching IDs
        """

        def check_id(id_: str, class_: str) -> bool:
            result = id_[: id_.find(ID_SPLIT_SYMBOL)] == class_

            if result and key is not None:
                result = id_[id_.rfind(ID_SPLIT_SYMBOL) + 1 :] == key
            return result

        # Define spaces to search
        spaces = {
            ExperimentDataEnum.additional_fields: self.additional_fields.columns,
            ExperimentDataEnum.analysis_tables: self.analysis_tables.keys(),
            ExperimentDataEnum.groups: self.groups.keys(),
            ExperimentDataEnum.variables: self.variables.keys(),
        }

        # Convert classes to names
        classes = [
            c.__name__ if isinstance(c, type) else c for c in Adapter.to_list(classes)
        ]

        # Get spaces to search
        searched_space = (
            Adapter.to_list(searched_space) if searched_space else list(spaces.keys())
        )

        # Return matching IDs
        return {
            class_: {
                space.value: [
                    str(id_) for id_ in spaces[space] if check_id(id_, class_)
                ]
                for space in searched_space
            }
            for class_ in classes
        }

    def get_one_id(
        self,
        class_: Union[type, str],
        space: ExperimentDataEnum,
        key: Optional[str] = None,
    ) -> str:
        """
        Get a single ID matching criteria.

        Args:
            class_: Class to search for
            space: Space to search in
            key: Optional key to match

        Returns:
            str: Matching ID

        Raises:
            NotFoundInExperimentDataError: If no matching ID found
        """
        class_ = class_ if isinstance(class_, str) else class_.__name__
        result = self.get_ids(class_, space, key)
        if (class_ not in result) or (not len(result[class_][space.value])):
            raise NotFoundInExperimentDataError(class_)
        return result[class_][space.value][0]

    def copy(self, data: Optional[Dataset] = None) -> "ExperimentData":
        """
        Create a deep copy.

        Args:
            data: Optional new base dataset

        Returns:
            ExperimentData: Deep copy of self
        """
        result = deepcopy(self)
        if data is not None:
            result._data = data
        return result

    def field_search(
        self,
        roles: Union[ABCRole, Iterable[ABCRole]],
        tmp_role: bool = False,
        search_types=None,
    ) -> List[str]:
        """
        Search for fields with specified roles.

        Args:
            roles: Roles to search for
            tmp_role: Whether to include temporary roles
            search_types: Types to search for

        Returns:
            List[str]: Matching field names
        """
        searched_field = []
        roles = Adapter.to_list(roles)

        # Split roles by type
        field_in_additional = [
            role for role in roles if isinstance(role, AdditionalRole)
        ]
        field_in_data = [role for role in roles if role not in field_in_additional]

        # Search in main data
        if field_in_data:
            searched_field += self.ds.search_columns(
                field_in_data, tmp_role=tmp_role, search_types=search_types
            )

        # Search in additional fields
        if field_in_additional and isinstance(self, ExperimentData):
            searched_field += self.additional_fields.search_columns(
                field_in_additional, tmp_role=tmp_role, search_types=search_types
            )

        return searched_field

    def field_data_search(
        self,
        roles: Union[ABCRole, Iterable[ABCRole]],
        tmp_role: bool = False,
        search_types=None,
    ) -> Dataset:
        """
        Search for data with specified roles.

        Args:
            roles: Roles to search for
            tmp_role: Whether to include temporary roles
            search_types: Types to search for

        Returns:
            Dataset: Dataset containing matching fields
        """
        searched_data: Dataset = Dataset.create_empty()
        roles = Adapter.to_list(roles)

        # Map roles to columns
        roles_columns_map = {
            role: self.field_search(role, tmp_role, search_types) for role in roles
        }

        # Build dataset from matching columns
        for role, columns in roles_columns_map.items():
            for column in columns:
                t_data = (
                    self.additional_fields[column]
                    if isinstance(role, AdditionalRole)
                    else self.ds[column]
                )
                searched_data = searched_data.add_column(
                    data=t_data, role={column: role}
                )
        return searched_data


class DatasetAdapter(Adapter):
    """
    Adapter class for converting various data types to Dataset objects.
    """

    @staticmethod
    def to_dataset(
        data: Union[Dict, Dataset, pd.DataFrame, List, str, int, float, bool],
        roles: Union[ABCRole, Dict[str, ABCRole]],
    ) -> Dataset:
        """
        Convert various data types to a Dataset object.

        Args:
            data: Input data to convert
            roles: Roles for the dataset columns

        Returns:
            Dataset: Converted dataset

        Raises:
            InvalidArgumentError: If data type is not supported
        """
        # Convert data based on its type
        if isinstance(data, Dict):
            return DatasetAdapter.dict_to_dataset(data, roles)
        elif isinstance(data, pd.DataFrame):
            if isinstance(roles, ABCRole):
                raise InvalidArgumentError("roles", "Dict[str, ABCRole]")
            return DatasetAdapter.frame_to_dataset(data, roles)
        elif isinstance(data, List):
            if isinstance(roles, ABCRole):
                raise InvalidArgumentError("roles", "Dict[str, ABCRole]")
            return DatasetAdapter.list_to_dataset(data, roles)
        elif any(isinstance(data, t) for t in [str, int, float, bool]):
            return DatasetAdapter.value_to_dataset(data, roles)
        elif isinstance(data, Dataset):
            return data
        else:
            raise InvalidArgumentError("data", "Dict, pd.DataFrame, List, Dataset")

    @staticmethod
    def value_to_dataset(
        data: ScalarType, roles: Union[ABCRole, Dict[str, ABCRole]]
    ) -> Dataset:
        """
        Convert a scalar value to a Dataset.

        Args:
            data: Scalar value to convert
            roles: Roles for the dataset

        Returns:
            Dataset: Dataset containing the scalar value
        """
        if isinstance(roles, ABCRole):
            roles = {"value": roles}
        return Dataset(roles=roles, data=pd.DataFrame({list(roles.keys())[0]: [data]}))

    @staticmethod
    def dict_to_dataset(
        data: Dict, roles: Union[ABCRole, Dict[str, ABCRole]]
    ) -> Dataset:
        """
        Convert a dictionary to a Dataset.

        Args:
            data: Dictionary to convert
            roles: Roles for the dataset

        Returns:
            Dataset: Dataset created from dictionary
        """
        roles_names = list(data.keys())
        if any(
            [
                any(isinstance(i, t) for t in [int, str, float, bool])
                for i in list(data.values())
            ]
        ):
            data = [data]
        if isinstance(roles, Dict):
            return Dataset.from_dict(data=data, roles=roles)
        elif isinstance(roles, ABCRole):
            return Dataset.from_dict(
                data=data, roles={name: roles for name in roles_names}
            )

    @staticmethod
    def list_to_dataset(data: List, roles: Dict[str, ABCRole]) -> Dataset:
        """
        Convert a list to a Dataset.

        Args:
            data: List to convert
            roles: Roles for the dataset

        Returns:
            Dataset: Dataset created from list
        """
        return Dataset(
            roles=roles,
            data=pd.DataFrame(data=data, columns=[list(roles.keys())[0]]),
        )

    @staticmethod
    def frame_to_dataset(data: pd.DataFrame, roles: Dict[str, ABCRole]) -> Dataset:
        """
        Convert a pandas DataFrame to a Dataset.

        Args:
            data: DataFrame to convert
            roles: Roles for the dataset

        Returns:
            Dataset: Dataset created from DataFrame
        """
        return Dataset(
            roles=roles,
            data=data,
        )
